"""Authentication flow and session management."""
from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from functools import cached_property
import json
import logging
import os
from pathlib import Path
import threading
import time
from typing import TYPE_CHECKING, Dict, Final, List, Optional, Tuple, Union, cast
import webbrowser

import jwt
import requests
from requests.exceptions import InvalidJSONError

from bitfount import web_utils
from bitfount.config import (
    _DEVELOPMENT_ENVIRONMENT,
    _STAGING_ENVIRONMENT,
    BITFOUNT_STORAGE_PATH,
    _get_environment,
)
from bitfount.exceptions import BitfountError
from bitfount.hub.types import (
    _DEV_AM_URL,
    _DEV_HUB_URL,
    _STAGING_AM_URL,
    _STAGING_HUB_URL,
    PRODUCTION_AM_URL,
    PRODUCTION_HUB_URL,
    APIKeys,
    _DeviceAccessTokenFailResponseJSON,
    _DeviceAccessTokenRequestDict,
    _DeviceAccessTokenResponseJSON,
    _DeviceCodeRequestDict,
    _DeviceCodeResponseJSON,
    _TokenRefreshRequestDict,
    _TokenRefreshResponseJSON,
)
from bitfount.utils import delegates
from bitfount.web_utils import _auto_retry_request

if TYPE_CHECKING:
    from requests import Response

logger = logging.getLogger(__name__)

# This forces `requests` to make IPv4 connections
# TODO: [BIT-1443] Remove this once Hub/AM support IPv6
requests.packages.urllib3.util.connection.HAS_IPV6 = False  # type: ignore[attr-defined] # Reason: see above # noqa: B950

_HUB_URLS: List[str] = [PRODUCTION_HUB_URL, _STAGING_HUB_URL, _DEV_HUB_URL]
_AM_URLS: List[str] = [PRODUCTION_AM_URL, _STAGING_AM_URL, _DEV_AM_URL]


_PRODUCTION_AUTH_DOMAIN: Final = "auth.bitfount.com"
_STAGING_AUTH_DOMAIN: Final = "auth.staging.bitfount.com"
_DEVELOPMENT_AUTH_DOMAIN: Final = (
    "auth.staging.bitfount.com"  # this is part of the staging tenant on auth0
)

# TODO: [BIT-356] potentially remove these client ids from the codebase
_PRODUCTION_CLIENT_ID: Final = "8iCJ33Kp6hc9ofrXTzr5GLxMRHWrlzZO"
_STAGING_CLIENT_ID: Final = "Wk4XZHDKfY8F3OYcKdagIHETt6JYwX08"
_DEVELOPMENT_CLIENT_ID: Final = "MP8oao6gcJd4jARwzJiJlEiK59ZeLCt3"

_SCOPES: Final = "profile openid offline_access"
_HUB_API_IDENTIFIER: Final = (
    "https://hub.bitfount.com/api"  # this is the same for staging and production
)
_DEVICE_CODE_GRANT_TYPE: Final = "urn:ietf:params:oauth:grant-type:device_code"
_AUTHORIZATION_PENDING_ERROR: Final = "authorization_pending"
_SLOW_DOWN_ERROR: Final = "slow_down"

_DEFAULT_USERNAME: Final[str] = "_default"
_USERNAME_KEY: Final = "https://www.bitfount.com/username"


@dataclass
class _AuthEnv:
    """Captures the combined authorisation information."""

    name: str
    auth_domain: str
    client_id: str


def _get_auth_environment() -> _AuthEnv:
    """Determines the auth settings based on environment variables.

    Returns:
        A tuple containing the auth domain and client ID for the given environment.
    """
    environment = _get_environment()
    if environment == _STAGING_ENVIRONMENT:
        return _AuthEnv("staging", _STAGING_AUTH_DOMAIN, _STAGING_CLIENT_ID)
    if environment == _DEVELOPMENT_ENVIRONMENT:
        return _AuthEnv("development", _DEVELOPMENT_AUTH_DOMAIN, _DEVELOPMENT_CLIENT_ID)
    return _AuthEnv("production", _PRODUCTION_AUTH_DOMAIN, _PRODUCTION_CLIENT_ID)


class AuthEnvironmentError(BitfountError):
    """Exception related to the authorization and authentication environment."""

    pass


@delegates()
class BitfountSession(requests.Session):
    """Manages session-based interactions with Bitfount.

    Extends requests.Session, appending an access token to the
    authorization of any requests made if an access token is present

    When the token expires it will request a new token prior to
    sending the web request.

    Usage:
        `session = BitfountSession(...)`
        # When you want the user to interact to start the session:
        `session.authenticate()`
        # The session can then be used as a normal requests.Session


    Attributes:
        access_token_expires_at: The time at which the access token expires.
        device_code: The device code returned by the Bitfount API.
        device_code_arrival_time: The time at which the device code was issued.
        device_code_expires_at: The time at which the device code expires.
        id_token: The ID token returned by the Bitfount API.
        refresh_token: The refresh token returned by the Bitfount API.
        token_file: The path to the file where the token is stored.
        token_request_interval: The time between token requests.
        user_storage_path: The path to the file where the user is stored.
    """

    def __init__(
        self,
        auth_domain: str = _PRODUCTION_AUTH_DOMAIN,
        client_id: str = _PRODUCTION_CLIENT_ID,
        username: str = _DEFAULT_USERNAME,
        scopes: str = _SCOPES,
        audience: str = _HUB_API_IDENTIFIER,
    ):
        super().__init__()

        self._auth_domain = auth_domain
        self._client_id = client_id
        self._username = username
        self._scopes: str = scopes
        self._audience: str = audience
        self._device_code_endpoint: str = (
            f"https://{self._auth_domain}/oauth/device/code"
        )
        self._token_endpoint: str = f"https://{self._auth_domain}/oauth/token"
        self._access_token: Optional[str] = None
        self._reauthentication_lock = threading.Lock()

        self.access_token_expires_at: Optional[datetime] = None
        self.device_code: Optional[str] = None
        self.device_code_arrival_time: Optional[datetime] = None
        self.device_code_expires_in: Optional[int] = None
        self.id_token: Optional[str] = None
        self.refresh_token: Optional[str] = None
        self.user_storage_path: Path = BITFOUNT_STORAGE_PATH / username
        self.token_file: Path = self.user_storage_path / ".token"
        self.token_request_interval: Optional[int] = None

    @cached_property
    def api_keys(self) -> Optional[APIKeys]:
        """Returns API keys to be used for the current session if they exist.

        These are cached so they don't need to be retrieved every time.
        """
        if (access_key_id := os.environ.get("BITFOUNT_API_KEY_ID")) and (
            access_key := os.environ.get("BITFOUNT_API_KEY")
        ):
            logger.info("Using API keys for authentication.")
            return APIKeys(access_key_id=access_key_id, access_key=access_key)

        return None

    @property
    def username(self) -> str:
        """Returns the username of the authenticated user."""
        if self._username != _DEFAULT_USERNAME:
            return self._username

        return self._get_username_from_id_token()

    @property
    def access_token(self) -> Optional[str]:
        """Returns the access token. Tries to refresh it if needed."""
        with self._reauthentication_lock:
            if not self.authenticated:
                self.authenticate()

            return self._access_token

    @property
    def authenticated(self) -> bool:
        """Returns true if we have an unexpired access token or API Keys."""
        return bool(self.api_keys) or bool(self._access_token and not self.expired)

    @property
    def expired(self) -> bool:
        """Whether or not the access token has expired.

        Returns: True if the token has expired
        """
        # Either both attributes will be present or neither will be
        if self._access_token and self.access_token_expires_at:
            # If the token expires in the next 10 minutes, refresh
            return (
                self.access_token_expires_at - timedelta(minutes=10)
            ) <= datetime.now(timezone.utc)
        else:
            return False

    @staticmethod
    def _is_url_in_urls(url: str, urls: List[str]) -> bool:
        """Returns true if the given `url` is in the list of `urls`.

        This includes if `url` points to a particular resource/page/endpoint etc. for
        a url present in `urls`.
        """
        for _url in urls:
            if url.startswith(_url):
                return True

        return False

    @classmethod
    def _is_hub_url(cls, url: str) -> bool:
        """Returns whether the provided url is a Bitfount Hub URL."""
        return cls._is_url_in_urls(url, _HUB_URLS)

    @classmethod
    def _is_am_url(cls, url: str) -> bool:
        """Returns whether the provided url is a Bitfount AM URL."""
        return cls._is_url_in_urls(url, _AM_URLS)

    def _verify_user_storage_path(self) -> None:
        """Verifies that user storage path corresponds to username.

        Raises:
            AssertionError: if user storage path corresponds to a different username
                from the BitfountSession.
        """
        if not str(self.user_storage_path).endswith((_DEFAULT_USERNAME, self.username)):
            provided_user = self.user_storage_path.stem
            raise AssertionError(
                f"BitfountSession connected to {self.username}. "
                f"This does not match the provided user storage path: {provided_user}"
            )

    def authenticate(self) -> None:
        """Authenticates user to allow protected requests.

        Prompts the user to login/authenticate and stores the tokens to use them
        in future requests.

        Raises:
            AssertionError: If user storage path corresponds to a different username
                from the BitfountSession.
            ConnectionError: If a token cannot be retrieved.
        """
        if self.api_keys:
            logger.debug("Using API keys, no need to authenticate.")
            return

        self._load_token_from_file(self.token_file)

        # Refresh the loaded token if it has expired
        refreshed = False
        if self.expired:
            refreshed = self._refresh_access_token()

        # Force user to go through login flow if we didn't refresh the token
        # Or if we haven't loaded an authenticated token
        if not self.authenticated and not refreshed:
            user_code, verification_uri = self._fetch_device_code()
            self._do_verification(user_code, verification_uri)
            self._exchange_device_code_for_token()

        # Verify that user storage path corresponds to username before saving the token
        self._verify_user_storage_path()

        # Ensure directory path exists
        self.user_storage_path.mkdir(parents=True, exist_ok=True)
        self._save_token_to_file(self.token_file)
        logger.info(f'Logged into Bitfount as "{self.username}"')

    def _fetch_device_code(self) -> Tuple[str, str]:
        """Fetches device code."""
        # See: https://auth0.com/docs/api/authentication?http#device-authorization-flow
        request_data: _DeviceCodeRequestDict = {
            "client_id": self._client_id,
            "scope": self._scopes,
            "audience": self._audience,
        }
        device_code_response: Response = web_utils.post(
            self._device_code_endpoint,
            data=request_data,
        )
        device_code_response.raise_for_status()

        response_json: _DeviceCodeResponseJSON = device_code_response.json()
        verification_uri: str = response_json["verification_uri_complete"]
        user_code: str = response_json["user_code"]

        self.device_code = response_json["device_code"]
        self.token_request_interval = response_json["interval"]
        self.device_code_expires_in = response_json["expires_in"]

        # This doesn't need to be exact, network latency affects this anyway
        self.device_code_arrival_time = datetime.now(timezone.utc)

        return user_code, verification_uri

    def _do_verification(self, user_code: str, verification_uri: str) -> None:
        """Opens web browser for verification."""
        print(f"Your confirmation code is: {user_code}")
        time.sleep(1)  # Give the user a second to see the code before opening browser
        webbrowser.open(verification_uri)
        print(
            "A browser window has been opened, "
            "please log in to confirm your identity."
        )
        print("If no browser window has opened, then please visit the following URL:")
        print(verification_uri)

    def _exchange_device_code_for_token(self) -> None:
        """Exchanges device code for token."""
        token_response: Optional[
            Union[_DeviceAccessTokenResponseJSON, _DeviceAccessTokenFailResponseJSON]
        ] = None

        # This method should only be called after a call to _fetch_device_code
        # so these will be set. Asserts to reassure mypy.
        assert self.device_code_arrival_time is not None  # nosec[assert_used]
        assert self.device_code_expires_in is not None  # nosec[assert_used]
        assert self.token_request_interval is not None  # nosec[assert_used]

        interval = self.token_request_interval

        while not self._device_code_expired(
            self.device_code_arrival_time, self.device_code_expires_in
        ):
            response: Response = self._send_token_request()

            # Break out of loop as we have our tokens!
            if (status_code := response.status_code) == 200:
                try:
                    token_response = cast(
                        _DeviceAccessTokenResponseJSON, response.json()
                    )
                except InvalidJSONError:
                    logger.error(
                        f"Received 200 status response, but JSON is invalid: "
                        f'"{response.text}"'
                    )
                    pass
                # Break because we have token or because we're unable to decode it
                break

            # Treat it as an expected "failure" response until we know otherwise;
            # status code could be any 4XX value, so we instead just check for the
            # right format and error values.
            try:
                token_response = cast(
                    _DeviceAccessTokenFailResponseJSON, response.json()
                )
            except InvalidJSONError:
                logger.error(
                    f"Received {status_code} status response, but JSON is invalid: "
                    f'"{response.text}"'
                )
                break

            # Break out of loop unless the flow is still in progress
            if (error := token_response.get("error")) not in (
                _AUTHORIZATION_PENDING_ERROR,
                _SLOW_DOWN_ERROR,
            ):
                # Not a retry-able response; fail out
                error_msg = (
                    f"An unexpected error occurred: status code: {status_code}; "
                    f'"{response.text}"'
                )
                logger.error(error_msg)
                print(error_msg)
                break
            elif error == _SLOW_DOWN_ERROR:
                # Somehow polling too fast (though should be using the interval
                # they specified); increase interval
                logger.warning(
                    f"Polling too quickly; increasing interval from {interval} "
                    f"to {interval+1} seconds"
                )
                interval += 1
            else:  # error == _AUTHORIZATION_PENDING_ERROR
                # Fine and expected, just keep trying
                pass

            print(
                f"Awaiting authentication in browser. "
                f"Will check again in {interval} seconds."
            )
            time.sleep(interval)

        if token_response and "access_token" in token_response:
            # Have our response now
            token_response = cast(_DeviceAccessTokenResponseJSON, token_response)
            self._access_token = token_response["access_token"]
            self.refresh_token = token_response["refresh_token"]
            self.id_token = token_response["id_token"]
            self.access_token_expires_at = datetime.now(timezone.utc) + timedelta(
                seconds=token_response["expires_in"],
            )
        else:
            raise ConnectionError(
                "Failed to retrieve a device code from the authentication server"
            )

    def _send_token_request(self, refresh: bool = False) -> Response:
        """Sends a request to the Auth Server token endpoint to get a new token."""
        if refresh:
            # See: https://auth0.com/docs/api/authentication?http#refresh-token
            # If refreshing, must have refresh_token. Reassure mypy.
            assert self.refresh_token is not None  # nosec[assert_used]
            refresh_request_data: _TokenRefreshRequestDict = {
                "client_id": self._client_id,
                "grant_type": "refresh_token",
                "refresh_token": self.refresh_token,
            }
            return web_utils.post(self._token_endpoint, data=refresh_request_data)
        else:
            # See: https://auth0.com/docs/api/authentication?http#device-authorization-flow48  # noqa: B950
            # device_code must already be set by the time we are calling this.
            # Reassure mypy.
            assert self.device_code is not None  # nosec[assert_used]
            request_data: _DeviceAccessTokenRequestDict = {
                "client_id": self._client_id,
                "grant_type": _DEVICE_CODE_GRANT_TYPE,
                "device_code": self.device_code,
            }
            return web_utils.post(self._token_endpoint, data=request_data)

    # We only wrap this method in _auto_retry_request as any calls to the others
    # (post, get, etc) will make use of this. Wrapping them all would result in
    # a double retry loop, but we can't _not_ wrap request as it is often used
    # directly.
    @_auto_retry_request
    def request(  # type: ignore[no-untyped-def,override]
        self, method, url, params=None, data=None, headers=None, **kwargs
    ) -> Response:
        """Performs an HTTP request.

        Overrides requests.session.request, appending our access token
        to the request headers or API keys if present.
        """
        # Create headers if they don't exist already
        if not headers:
            headers = {}

        is_am_url = self._is_am_url(url)
        is_hub_url = self._is_hub_url(url)

        if self.api_keys and (is_am_url or is_hub_url):
            logger.debug(f"Adding API key to request headers for {url}")

            if is_hub_url:
                # If the URL is a hub URL, only the first half of the API keys are used
                # which corresponds to the portion required for the Hub
                headers["x-api-key-id"] = self.api_keys.access_key_id.split(":")[0]
                headers["x-api-key"] = self.api_keys.access_key.split(":")[0]

            elif is_am_url:
                # If the URL is an AM URL, the entire API key is used because the AM
                # also calls the Hub under the hood and therefore needs both portions
                headers["x-api-key-id"] = self.api_keys.access_key_id
                headers["x-api-key"] = self.api_keys.access_key

        elif self.access_token:
            logger.debug(f"Adding token to request headers for {url}")
            headers["authorization"] = f"Bearer {self._access_token}"

        return super().request(
            method, url, params=params, data=data, headers=headers, **kwargs
        )

    def _refresh_access_token(self) -> bool:
        """Attempts to refresh the access token.

        Returns: True if the token was refreshed, false otherwise

        """
        token_response = self._send_token_request(refresh=True)

        if token_response.status_code == 200:
            token_response_json: _TokenRefreshResponseJSON = token_response.json()
            self._access_token = token_response_json["access_token"]
            self.refresh_token = token_response_json["refresh_token"]
            self.id_token = token_response_json["id_token"]
            self.access_token_expires_at = datetime.now(timezone.utc) + timedelta(
                seconds=token_response_json["expires_in"]
            )
            return True
        logger.warning(
            f"Failed to refresh access token, response was: {token_response.text}"
        )
        return False

    @staticmethod
    def _device_code_expired(arrival_time: datetime, expires_in_seconds: int) -> bool:
        """Checks if authorization code has expired.

        Checks if too much time has passed between the device code
        being issued by the Auth Server and the user approving access
        using that device code.
        """
        return datetime.now(timezone.utc) >= arrival_time + timedelta(
            seconds=expires_in_seconds
        )

    def _save_token_to_file(self, token_file: Path) -> None:
        """Saves authentication token to file.

        Saves all fields that are necessary to reproduce this object to a file.
        """
        # This will be set by the time this is called. Reassure mypy.
        assert self.access_token_expires_at is not None  # nosec[assert_used]
        json.dump(
            {
                "access_token": self._access_token,
                "refresh_token": self.refresh_token,
                "id_token": self.id_token,
                "access_token_expires_at": self.access_token_expires_at.timestamp(),
                "auth_domain": self._auth_domain,
                "client_id": self._client_id,
                "scopes": self._scopes,
                "audience": self._audience,
            },
            token_file.open("w"),
        )

    def _load_token_from_file(self, token_file: Path) -> None:
        """Loads authentication token from file.

        Attempts to load the data needed for authentication, when loaded it updates
        the fields on the instance.

        If the data is found but the metadata differs then it will not update
        the fields.

        If no data, or no file is found it will just return without error.
        """
        if token_file.exists():
            serialized_tokens = json.load(token_file.open())

            auth_domain = serialized_tokens["auth_domain"]
            client_id = serialized_tokens["client_id"]
            scopes = serialized_tokens["scopes"]
            audience = serialized_tokens["audience"]

            if (
                self._auth_domain != auth_domain
                or self._client_id != client_id
                or self._scopes != scopes
                or self._audience != audience
            ):
                print(
                    "Stored tokens are no longer valid, "
                    "fresh authentication is necessary"
                )
                return

            self._access_token = serialized_tokens["access_token"]
            self.refresh_token = serialized_tokens["refresh_token"]
            self.id_token = serialized_tokens["id_token"]
            self.access_token_expires_at = datetime.fromtimestamp(
                serialized_tokens["access_token_expires_at"], tz=timezone.utc
            )

    def _get_username_from_id_token(self) -> str:
        """Extracts the Bitfount username from the token.

        Note: This function performs no verification of the id_token signature
        and should only be used in situations where the username in the token
        is not used to make decisions. As this is not backend code (i.e.
        anyone can edit this) we aren't very concerned about the fact it is
        not verified.
        """
        # Calling function assures authenticated. Reassure mypy.
        assert self.id_token is not None  # nosec[assert_used]
        # Decode the ID token without verification
        id_token_decode: Dict[str, str] = jwt.decode(
            self.id_token, options={"verify_signature": False}
        )
        return id_token_decode[_USERNAME_KEY]



"""Module containing BaseSource class.

BaseSource is the abstract data source class from which all concrete data sources
must inherit.
"""
from __future__ import annotations

from abc import ABC, abstractmethod
import inspect
import logging
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    TypeVar,
    Union,
    cast,
)

import methodtools
from mypy_extensions import Arg, DefaultNamedArg, KwArg, VarArg
import numpy as np
import pandas as pd
from pandas.core.dtypes.common import is_datetime64_any_dtype as is_datetime

from bitfount.data.datasources.utils import _modify_column, _modify_file_paths
from bitfount.data.datasplitters import DatasetSplitter
from bitfount.data.exceptions import DataNotLoadedError
from bitfount.data.types import (
    DataPathModifiers,
    _Column,
    _GetColumnCallable,
    _GetDtypesCallable,
)
from bitfount.data.utils import _generate_dtypes_hash, _hash_str
from bitfount.types import _Dtypes
from bitfount.utils import delegates, seed_all

logger = logging.getLogger(__name__)


T = TypeVar("T", bound="BaseSource")
BaseSourceInitSignature = Callable[
    [
        Arg(T, "self"),  # noqa: F821
        VarArg(Any),
        DefaultNamedArg(Optional[DatasetSplitter], "data_splitter"),  # noqa: F821
        DefaultNamedArg(Optional[int], "seed"),  # noqa: F821
        DefaultNamedArg(
            Optional[Dict[str, DataPathModifiers]], "modifiers"  # noqa: F821
        ),
        DefaultNamedArg(Union[str, Sequence[str], None], "ignore_cols"),  # noqa: F821
        KwArg(Any),
    ],
    None,
]


class BaseSource(ABC):
    """Abstract Base Source from which all other data sources must inherit.

    Args:
        data_splitter: Approach used for splitting the data into training, test,
            validation. Defaults to None.
        seed: Random number seed. Used for setting random seed for all libraries.
            Defaults to None.
        modifiers: Dictionary used for modifying paths/ extensions in the dataframe.
            Defaults to None.
        ignore_cols: Column/list of columns to be ignored from the data.
            Defaults to None.

    Attributes:
        data: A Dataframe-type object which contains the data.
        data_splitter: Approach used for splitting the data into training, test,
            validation.
        seed: Random number seed. Used for setting random seed for all libraries.
        train_idxs: A numpy array containing the indices of the data which
            will be used for training.
        validation_idxs: A numpy array containing the indices of the data which
            will be used for validation.
        test_idxs: A numpy array containing the indices of the data which
            will be used for testing.

    """

    def __init__(
        self,
        data_splitter: Optional[DatasetSplitter] = None,
        seed: Optional[int] = None,
        modifiers: Optional[Dict[str, DataPathModifiers]] = None,
        ignore_cols: Optional[Union[str, Sequence[str]]] = None,
        **kwargs: Any,
    ) -> None:
        self._base_source_init = True
        self.data_splitter = data_splitter
        self.seed = seed
        self._modifiers = modifiers
        self._data_is_split: bool = False
        self._data_is_loaded: bool = False
        seed_all(self.seed)

        self._train_idxs: Optional[np.ndarray] = None
        self._validation_idxs: Optional[np.ndarray] = None
        self._test_idxs: Optional[np.ndarray] = None

        self._data: pd.DataFrame
        self._table_hashes: Set[str] = set()

        self._ignore_cols: List[str] = []
        if isinstance(ignore_cols, str):
            self._ignore_cols = [ignore_cols]
        elif ignore_cols is not None:
            self._ignore_cols = list(ignore_cols)

    def __init_subclass__(cls, **kwargs: Any) -> None:
        if not (inspect.isabstract(cls) or ABC in cls.__bases__):
            cls.get_dtypes = cls._get_dtypes()  # type: ignore[assignment] # reason: wrap subclass get_dtypes # noqa: B950
            cls.get_column = cls._get_column()  # type: ignore[assignment] # reason: wrap subclass get_column # noqa: B950

    @classmethod
    def _get_dtypes(cls) -> _GetDtypesCallable:
        """Decorate subclass' get_dtypes implementation.

        Decorate subclass' implementation of get_dtypes to handle ignored
        columns and handle `_table_hashes`.
        """
        subclass_get_dtypes = cls.get_dtypes

        def get_dtypes(self: BaseSource, *args: Any, **kwargs: Any) -> _Dtypes:
            dtypes: _Dtypes = subclass_get_dtypes(self, *args, **kwargs)
            if self._ignore_cols:
                for col in self._ignore_cols:
                    if col in dtypes:
                        del dtypes[col]
            self._table_hashes.add(_generate_dtypes_hash(dtypes))
            return dtypes

        return get_dtypes

    @classmethod
    def _get_column(
        cls,
    ) -> _GetColumnCallable:
        """Decorate subclass' get_column implementation.

        Decorate subclass' implementation of get_column to handle ignored
        columns and modifiers.
        """
        subclass_get_column = cls.get_column

        def get_column(
            self: BaseSource, col_name: str, *args: Any, **kwargs: Any
        ) -> _Column:

            column = subclass_get_column(self, col_name, *args, **kwargs)
            if self._modifiers:
                if modifier_dict := self._modifiers.get(col_name):
                    column = _modify_column(column, modifier_dict)
            return column

        return get_column

    @property
    def is_initialised(self) -> bool:
        """Checks if `BaseSource` was initialised."""
        if hasattr(self, "_base_source_init"):
            return True
        else:
            return False

    @property
    def data(self) -> pd.DataFrame:
        """Data."""
        if hasattr(self, "_data"):
            return self._data
        else:
            raise DataNotLoadedError(
                "Data is not loaded yet. Please call `load_data` first."
            )

    @data.setter
    def data(self, _data: Optional[pd.DataFrame]) -> None:
        """Data setter."""
        if _data is not None:
            if isinstance(_data, pd.DataFrame):

                if self._ignore_cols:
                    # If columns already ignored in data, ignore errors.
                    _data = _data.drop(columns=self._ignore_cols, errors="ignore")
                self._data = _data

                if self._modifiers:
                    _modify_file_paths(self._data, self._modifiers)

                self._data_is_loaded = True
            else:
                raise TypeError(
                    "Invalid data attribute. "
                    "Expected pandas dataframe for attribute 'data' "
                    f"but received :{type(_data)}"
                )

    @property
    def hash(self) -> str:
        """The hash associated with this BaseSource.

        This is the hash of the static information regarding the underlying DataFrame,
        primarily column names and content types but NOT anything content-related
        itself. It should be consistent across invocations, even if additional data
        is added, as long as the DataFrame is still compatible in its format.

        Returns:
            The hexdigest of the DataFrame hash.
        """
        if not self._table_hashes:
            raise DataNotLoadedError(
                "Data is not loaded yet. Please call `get_dtypes` first."
            )
        else:
            return _hash_str(str(sorted(self._table_hashes)))

    @abstractmethod
    def get_values(
        self, col_names: List[str], **kwargs: Any
    ) -> Dict[str, Iterable[Any]]:
        """Implement this method to get distinct values from list of columns."""
        raise NotImplementedError

    @abstractmethod
    def get_column(self, col_name: str, **kwargs: Any) -> _Column:
        """Implement this method to get single column from dataset."""
        raise NotImplementedError

    # TODO: [BIT-1780] Simplify referencing data in BaseSource
    # We want to avoid recalculating but we dont want to cache more
    # than one result at a time to save memory
    @methodtools.lru_cache(maxsize=1)  # type: ignore[misc]  # Reason: methodtools untyped  # noqa: B950
    @abstractmethod
    def get_data(self, **kwargs: Any) -> Optional[pd.DataFrame]:
        """Implement this method to load and return dataset."""
        raise NotImplementedError

    @abstractmethod
    def get_dtypes(self, **kwargs: Any) -> _Dtypes:
        """Implement this method to get the columns and column types from dataset."""
        raise NotImplementedError

    @abstractmethod
    def __len__(self) -> int:
        """Implement this method to get the number of rows in the dataset."""
        raise NotImplementedError

    @property
    @abstractmethod
    def multi_table(self) -> bool:
        """Implement this method to define whether the data source is multi-table."""
        raise NotImplementedError

    @staticmethod
    def _get_data_dtypes(data: pd.DataFrame) -> _Dtypes:
        """Returns the nullable column types of the dataframe.

        This is called by the `get_dtypes` method. This method also overrides datetime
        column dtypes to be strings. This is not done for date columns which are of
        type object.
        """
        data = cast(pd.DataFrame, data.convert_dtypes())
        dtypes: _Dtypes = data.dtypes.to_dict()
        for name in list(dtypes):
            if is_datetime(data[name]):
                dtypes[name] = pd.StringDtype()

        return dtypes

    def load_data(self, **kwargs: Any) -> None:
        """Load the data for the datasource.

        We wrap get_data with lru_cache so this method is idempotent so it
        can be called multiple times with the same arguments without reloading
        the data.

        Raises:
            TypeError: If data format is not supported.
        """
        data = self.get_data(**kwargs)
        if data is not None:
            self.data = data


@delegates()
class MultiTableSource(BaseSource, ABC):
    """Abstract base source that supports multiple tables."""

    @property
    @abstractmethod
    def table_names(self) -> List[str]:
        """Implement this method to define whether the data source is multi-table."""
        raise NotImplementedError

    @abstractmethod
    def _validate_table_name(self, table_name: str) -> None:
        """Validate the table name exists in the multi-table datasource.

        This method should raise a ValueError if the table_name is not valid.
        """
        raise NotImplementedError

    @abstractmethod
    def get_values(
        self, col_names: List[str], table_name: Optional[str] = None, **kwargs: Any
    ) -> Dict[str, Iterable[Any]]:
        """Implement this method to get distinct values from list of columns."""
        raise NotImplementedError

    @abstractmethod
    def get_column(
        self, col_name: str, table_name: Optional[str] = None, **kwargs: Any
    ) -> _Column:
        """Implement this method to get single column from dataset."""
        raise NotImplementedError

    # TODO: [BIT-1780] Simplify referencing data in BaseSource
    # We want to avoid recalculating but we dont want to cache more
    # than one result at a time to save memory
    @methodtools.lru_cache(maxsize=1)  # type: ignore[misc]  # Reason: methodtools untyped  # noqa: B950
    @abstractmethod
    def get_data(
        self, table_name: Optional[str] = None, **kwargs: Any
    ) -> Optional[pd.DataFrame]:
        """Implement this method to loads and return dataset."""
        raise NotImplementedError

    @abstractmethod
    def get_dtypes(self, table_name: Optional[str] = None, **kwargs: Any) -> _Dtypes:
        """Implement this method to get the columns and column types from dataset."""
        raise NotImplementedError



"""Module containing CSVSource class.

CSVSource class handles loading of CSV data.
"""
from __future__ import annotations

import logging
import os
from typing import Any, Dict, Iterable, List, Optional, Union

import numpy as np
import pandas as pd
from pydantic import AnyUrl

from bitfount.data.datasources.base_source import BaseSource
from bitfount.types import _Dtypes
from bitfount.utils import delegates

logger = logging.getLogger(__name__)


@delegates()
class CSVSource(BaseSource):
    """Data source for loading csv files.

    Args:
        path: The path or URL to the csv file.
        read_csv_kwargs: Additional arguments to be passed as a
            dictionary to `pandas.read_csv`. Defaults to None.
    """

    def __init__(
        self,
        path: Union[os.PathLike, AnyUrl, str],
        read_csv_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        if not str(path).endswith(".csv"):
            raise TypeError("Please provide a Path or URL to a CSV file.")
        self.path = str(path)
        if not read_csv_kwargs:
            read_csv_kwargs = {}
        self.read_csv_kwargs = read_csv_kwargs

    def get_data(self, **kwargs: Any) -> pd.DataFrame:
        """Loads and returns data from CSV dataset.

        Returns:
            A DataFrame-type object which contains the data.
        """
        csv_df: pd.DataFrame = pd.read_csv(self.path, **self.read_csv_kwargs)
        return csv_df

    def get_values(
        self, col_names: List[str], **kwargs: Any
    ) -> Dict[str, Iterable[Any]]:
        """Get distinct values from columns in CSV dataset.

        Args:
            col_names: The list of the columns whose distinct values should be
                returned.

        Returns:
            The distinct values of the requested column as a mapping from col name to
            a series of distinct values.

        """
        return {col: self.get_data()[col].unique() for col in col_names}

    def get_column(self, col_name: str, **kwargs: Any) -> Union[np.ndarray, pd.Series]:
        """Loads and returns single column from CSV dataset.

        Args:
            col_name: The name of the column which should be loaded.

        Returns:
            The column request as a series.
        """
        csv_df: pd.DataFrame = self.get_data()
        return csv_df[col_name]

    def get_dtypes(self, **kwargs: Any) -> _Dtypes:
        """Loads and returns the columns and column types of the CSV dataset.

        Returns:
            A mapping from column names to column types.
        """
        csv_df: pd.DataFrame = self.get_data()
        return self._get_data_dtypes(csv_df)

    def __len__(self) -> int:
        return len(self.get_data())

    @property
    def multi_table(self) -> bool:
        """Attribute to specify whether the datasource is multi table."""
        return False



"""Classes for splitting data."""
from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
import logging
import random
from typing import TYPE_CHECKING, Any, List, Optional, Tuple, Union

import numpy as np
import pandas as pd

from bitfount.data.types import DataSplit

if TYPE_CHECKING:
    from bitfount.data.datasources.database_source import DatabaseSource


logger = logging.getLogger(__name__)


class DatasetSplitter(ABC):
    """Parent class for different types of dataset splits."""

    # TODO: [BIT-1053] If we're using python 3.9 we can stack property here too
    @classmethod
    @abstractmethod
    def splitter_name(cls) -> str:
        """Returns string name for splitter type."""
        raise NotImplementedError()

    @abstractmethod
    def create_dataset_splits(
        self, data: pd.DataFrame
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Returns indices for data sets."""
        raise NotImplementedError

    @abstractmethod
    def get_split_query(self, datasource: DatabaseSource, split: DataSplit) -> str:
        """Modifies the datasource SQL query to return a split of the data.

        Args:
            datasource: A `DatabaseSource` object.
            split: The relevant split to return from the query.

        Returns:
            The modified SQL query.
        """
        raise NotImplementedError

    @classmethod
    def create(cls, splitter_name: str, **kwargs: Any) -> DatasetSplitter:
        """Create a DataSplitter of the requested type."""
        # We may want to replace this with an `__init_subclass_` based
        # approach if we start adding more DataSplitters
        # See: https://blog.yuo.be/2018/08/16/__init_subclass__-a-simpler-way-to-implement-class-registries-in-python/ # noqa: B950
        if splitter_name == SplitterDefinedInData.splitter_name():
            return SplitterDefinedInData(**kwargs)
        return PercentageSplitter(**kwargs)


@dataclass
class PercentageSplitter(DatasetSplitter):
    """Splits data into sets based on percentages.

    The default split is 80% of the data is used training, and 10% for each
    validation and testing, respectively.

    Args:
        validation_percentage: The percentage of data to be used for validation.
            Defaults to 10.
        test_percentage: The percentage of data to be used for testing.
            Defaults to 10.
        time_series_sort_by: A string/list of strings to be used for sorting
            time series. The strings should correspond to feature names from the
            dataset. This sorts the dataframe by the values of those features
            ensuring the validation and test sets come after the training set data
            to remove potential bias during training and evaluation. Defaults to None.
    """

    validation_percentage: int = 10
    test_percentage: int = 10
    time_series_sort_by: Optional[Union[List[str], str]] = None

    def __post_init__(self) -> None:
        self.train_percentage = 100 - self.validation_percentage - self.test_percentage

    @classmethod
    def splitter_name(cls) -> str:
        """Class method for splitter name.

        Returns:
            The string name for splitter type.
        """
        return "percentage"

    def create_dataset_splits(
        self, data: pd.DataFrame
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Create splits in dataset for training, validation and test sets.

        Args:
            data: The dataframe type object to be split.

        Returns:
              A tuple of arrays, each containing the indices from the data
              to be used for training, validation, and testing, respectively.
        """
        # Sort or shuffle data depending on time_series constraint
        if self.time_series_sort_by:
            data = data.sort_values(by=self.time_series_sort_by)

        # Get indices of dataframe
        indices = data.index.to_series().to_list()

        if not self.time_series_sort_by:
            random.shuffle(indices)

        # Create dataset splits
        train_idxs, validation_idxs, test_idxs = np.split(
            indices,
            [
                int(
                    (100 - self.validation_percentage - self.test_percentage)
                    * len(indices)
                    / 100
                ),
                int((100 - self.test_percentage) * len(indices) / 100),
            ],
        )

        # Ensure that time series constraint is enforced correctly
        if self.time_series_sort_by:
            validation_idxs, test_idxs = self._split_indices_time_series(
                validation_idxs, test_idxs, data
            )
            train_idxs, validation_idxs = self._split_indices_time_series(
                train_idxs, validation_idxs, data
            )

            if self.validation_percentage == 0 and self.test_percentage != 0:
                train_idxs, test_idxs = self._split_indices_time_series(
                    train_idxs, test_idxs, data
                )

            np.random.shuffle(train_idxs)
            np.random.shuffle(validation_idxs)
            np.random.shuffle(test_idxs)

        return train_idxs, validation_idxs, test_idxs

    def get_split_query(self, datasource: DatabaseSource, split: DataSplit) -> str:
        """Modifies the datasource SQL query to return a split of the data.

        :::caution

        This method will only work for databases that support the `LIMIT ... OFFSET`
        syntax. Notably, Microsoft SQL Server does not support this syntax.

        :::

        :::caution

        It is strongly recommended that you sort the data as part of the SQL query in
        order to ensure the splits are random. This is because for iterable datasets,
        the splits are simply taken in order from TRAIN to TEST.

        Similarly, `time_series_sort_by` is ignored and a warning logged if True. If you
        want to sort by time series, you should do this as part of the SQL Query.

        :::

        Args:
            datasource: A `DatabaseSource` object.
            split: The relevant split to return from the query.

        Returns:
            The modified SQL query.
        """
        if self.time_series_sort_by:
            logger.warning(
                "Time series sort by is not supported for Database percentage splits. "
                "The sort by will be ignored. If you want to use time series sort by, "
                "please sort the dataset as you want in the SQL query."
            )

        len_datasource = len(datasource)

        if split == DataSplit.TRAIN:
            limit = int(len_datasource * self.train_percentage / 100)
            offset = 0
        elif split == DataSplit.VALIDATION:
            limit = int(len_datasource * self.validation_percentage / 100)
            offset = int(len_datasource * self.train_percentage / 100)
        elif split == DataSplit.TEST:
            limit = int(len_datasource * self.test_percentage / 100)
            offset = int(len_datasource * (100 - self.test_percentage) / 100)
        else:
            raise ValueError("Split not recognised")

        # Ignoring the security warning because the sql query is trusted and
        # will be executed regardless.
        return f"SELECT * FROM ({datasource.query}) q LIMIT {limit} OFFSET {offset}"  # nosec # noqa: B950

    def _split_indices_time_series(
        self, index_array_a: np.ndarray, index_array_b: np.ndarray, data: pd.DataFrame
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Splits indices in a chronological fashion.

        Helper function which ensures that two arrays of indices of the full
        dataset do not overlap in terms of dates to ensure there is no
        information leakage. This is necessary when there is not enough
        granularity in the dates provided yet we still want to make sure that
        e.g. all the samples in the test set definitely come after the
        validation set and so on. For example in the LendingClub data we are
        only given month and year - this function ensures that the
        training/validation/test splits only occur at month boundaries.
        Args:
            index_array_a: An array of indices from the full dataset.
            index_array_b: An array of indices from the full dataset.
            data: A dataframe type object.

        Returns:
             The two arrays of indices modified by moving items from
             index_list_a into index_list_b until the condition is satisfied.
        """
        # Get the most granular time series column
        time_series_col: Optional[str]
        if isinstance(self.time_series_sort_by, list):
            time_series_col = self.time_series_sort_by[-1]
        else:
            time_series_col = self.time_series_sort_by

        # Return if one of the arrays is empty
        if len(index_array_a) == 0 or len(index_array_b) == 0:
            return index_array_a, index_array_b

        first_val_b = data[time_series_col].loc[index_array_b[0].item()]
        last_val_a = data[time_series_col].loc[index_array_a[-1].item()]

        # Move data from a to b whilst condition is still satisfied
        while first_val_b == last_val_a:
            last_idx_a, index_array_a = index_array_a[-1], index_array_a[:-1]
            index_array_b = np.insert(index_array_b, 0, last_idx_a)

            if index_array_a.size == 0:
                raise ValueError(
                    "One or more of the training/test/validation sets ends "
                    "up empty in the _split_indices_time_series() function. "
                    "This is because one of the sets contains only one unique date."
                )
            first_val_b = data[time_series_col].loc[index_array_b[0].item()]
            last_val_a = data[time_series_col].loc[index_array_a[-1].item()]

        return index_array_a, index_array_b


@dataclass
class SplitterDefinedInData(DatasetSplitter):
    """Splits data into sets based on value in each row.

    The splitting is done based on the values in a user specified column.

    Args:
        column_name: The column name for which contains the labels
            for splitting. Defaults to "BITFOUNT_SPLIT_CATEGORY".
        training_set_label: The label for the data points to be included
            in the training set. Defaults to "TRAIN".
        validation_set_label: The label for the data points to be included
            in the validation set. Defaults to "VALIDATE".
        test_set_label: The label for the data points to be included in
            the test set. Defaults to "TEST".
    """

    column_name: str = "BITFOUNT_SPLIT_CATEGORY"
    training_set_label: str = "TRAIN"
    validation_set_label: str = "VALIDATE"
    test_set_label: str = "TEST"

    @classmethod
    def splitter_name(cls) -> str:
        """Class method for splitter name.

        Returns:
            The string name for splitter type.
        """
        return "predefined"

    def create_dataset_splits(
        self, data: pd.DataFrame
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Create splits in dataset for training, validation and test sets.

        Args:
            data: The dataframe type object to be split.

        Returns:
              A tuple of arrays, each containing the indices from the data
              to be used for training, validation, and testing, respectively.
        """
        training_indices = data[
            data[self.column_name] == self.training_set_label
        ].index.values
        validation_indices = data[
            data[self.column_name] == self.validation_set_label
        ].index.values
        test_indices = data[data[self.column_name] == self.test_set_label].index.values
        return training_indices, validation_indices, test_indices

    def get_split_query(self, datasource: DatabaseSource, split: DataSplit) -> str:
        """Modifies the datasource SQL query to return a split of the data.

        Args:
            datasource: A `DatabaseSource` object.
            split: The relevant split to return from the query.

        Returns:
            The modified SQL query.
        """
        if split == DataSplit.TRAIN:
            # Ignoring the security warning because the sql query is trusted and
            # will be executed regardless.
            query = (
                f"SELECT * FROM ({datasource.query}) q "  # nosec # Reason: See above
                f"WHERE {self.column_name} = '{self.training_set_label}'"
            )
        elif split == DataSplit.VALIDATION:
            query = (
                f"SELECT * FROM ({datasource.query}) q "  # nosec # Reason: See above
                f"WHERE {self.column_name} = '{self.validation_set_label}'"
            )
        elif split == DataSplit.TEST:
            query = (
                f"SELECT * FROM ({datasource.query}) q "  # nosec # Reason: See above
                f"WHERE {self.column_name} = '{self.test_set_label}'"
            )
        else:
            raise ValueError("Split not recognised")

        return query



"""Utility modules for tests."""
from _pytest.fixtures import SubRequest
from typing_extensions import TypeAlias

# It's actually `SubRequest` that has the `param` attribute guaranteed so this
# is what we want for type checking purposes, even though it's not part of the
# public API.
PytestRequest: TypeAlias = SubRequest



"""Algorithm to train and evaluate a model on remote data."""
from __future__ import annotations

from typing import Any, Dict, Mapping, cast

import numpy as np

from bitfount.data.datasources.base_source import BaseSource
from bitfount.federated.algorithms.model_algorithms.base import (
    _BaseModelAlgorithmFactory,
    _BaseModellerModelAlgorithm,
    _BaseWorkerModelAlgorithm,
)
from bitfount.federated.logging import _get_federated_logger
from bitfount.hub.api import BitfountHub
from bitfount.metrics import MetricCollection
from bitfount.utils import delegates

logger = _get_federated_logger(__name__)


class _ModellerSide(_BaseModellerModelAlgorithm):
    """Modeller side of the ModelTrainingAndEvaluation algorithm."""

    def run(
        self, results: Mapping[str, Dict[str, float]]
    ) -> Dict[str, Dict[str, float]]:
        """Simply returns results."""
        return dict(results)


class _WorkerSide(_BaseWorkerModelAlgorithm):
    """Worker side of the ModelTrainingAndEvaluation algorithm."""

    def run(self, data: BaseSource) -> Dict[str, float]:
        """Runs training and evaluation and returns metrics."""
        self.model.fit(data)
        preds, target = self.model.evaluate()
        # TODO: [BIT-1604] Remove these cast statements once they become superfluous.
        preds = cast(np.ndarray, preds)
        target = cast(np.ndarray, target)
        m = MetricCollection.create_from_model(self.model)
        return m.compute(target, preds)


@delegates()
class ModelTrainingAndEvaluation(_BaseModelAlgorithmFactory):
    """Algorithm for training a model, evaluating it and returning metrics.

    :::note

    The metrics cannot currently be specified by the user.

    :::

    Args:
        model: The model to train and evaluate on remote data.

    Attributes:
        model: The model to train and evaluate on remote data.
    """

    def modeller(self, **kwargs: Any) -> _ModellerSide:
        """Returns the modeller side of the ModelTrainingAndEvaluation algorithm."""
        model = self._get_model_from_reference()
        return _ModellerSide(model=model, **kwargs)

    def worker(self, hub: BitfountHub, **kwargs: Any) -> _WorkerSide:
        """Returns the worker side of the ModelTrainingAndEvaluation algorithm.

        Args:
            hub: `BitfountHub` object to use for communication with the hub.
        """
        model = self._get_model_from_reference(hub=hub)
        return _WorkerSide(model=model, **kwargs)



"""PyTorch backend test config settings."""
import inspect

from _pytest.monkeypatch import MonkeyPatch
from pytest import fixture

import bitfount.config


@fixture(autouse=True)
def env_fix(monkeypatch: MonkeyPatch) -> None:
    """Fix the environment into a known state for tests."""
    # Overrides the default fixture in tests/conftest.py
    monkeypatch.setenv("BITFOUNT_ENGINE", bitfount.config._PYTORCH_ENGINE)
    monkeypatch.setattr(
        "bitfount.config.BITFOUNT_ENGINE", bitfount.config._PYTORCH_ENGINE
    )


@fixture
def pytorch_bitfount_model_correct_structure() -> str:
    """Example good PytorchBitfountModel."""
    return inspect.cleandoc(
        """
        from torchmetrics.functional import accuracy
        import torch
        from torch import nn as nn
        from torch.nn import functional as F

        from bitfount.backends.pytorch.models.bitfount_model import (
            PyTorchBitfountModel,
        )
        from bitfount.backends.pytorch.models.base_models import (
            PyTorchClassifierMixIn,
        )

        class DummyModel(PyTorchClassifierMixIn, PyTorchBitfountModel):

            def __init__(self, epochs=5, **kwargs):
                super().__init__(epochs=epochs, **kwargs)
                self.learning_rate=0.01

            def create_model(self):
                self.input_size = self.datastructure.input_size

                return nn.Sequential(
                    nn.Linear(self.input_size, 500),
                    nn.ReLU(),
                    nn.Dropout(0.1),
                    nn.Linear(500, self.n_classes),
                )

            def forward(self, x):
                x, sup = x
                x = self._model(x.float())
                return x

            def training_step(self, batch, batch_idx):
                x, y = batch
                return F.cross_entropy(self(x), y)

            def validation_step(self, batch, batch_idx):
                x, y = batch
                preds = self(x)
                loss = F.cross_entropy(preds, y)
                preds = F.softmax(preds, dim=1)
                acc = accuracy(preds, y)

                # Calling self.log will surface up scalars for you in TensorBoard
                # self.log("val_loss", loss, prog_bar=True)
                # self.log("val_acc", acc, prog_bar=True)

                return {
                    "val_loss": loss,
                    "val_acc": acc,
                }

            def test_step(self, batch, batch_idx):
                x, y = batch
                preds = self(x)
                preds = F.softmax(preds, dim=1)

                # Output targets and prediction for later
                return {"predictions": preds, "targets": y}


            def configure_optimizers(self):
                optimizer = torch.optim.AdamW(self.parameters(), lr=self.learning_rate)
                return optimizer
        """
    )


@fixture
def pytorch_bitfount_model_tab_image_data() -> str:
    """Pytorch Model which handles both tabular and image data."""
    return inspect.cleandoc(
        """
        from collections import defaultdict
        from torchmetrics.functional import accuracy
        import torch
        from torch import nn as nn
        from torch.nn import functional as F

        from bitfount.backends.pytorch.models.bitfount_model import (
            PyTorchBitfountModel,
        )
        from bitfount.backends.pytorch.models.base_models import (
            PyTorchClassifierMixIn,
        )

        class DummyModelTabImg(PyTorchClassifierMixIn, PyTorchBitfountModel):

            def __init__(self, epochs=50, **kwargs):
                super().__init__(epochs=epochs, **kwargs)
                self.batch_size = 32
                self.learning_rate = 0.01
                self.num_workers = 0

            def create_model(self):

                class TabImg(nn.Module):

                    def __init__(self):
                        super().__init__()
                        self.conv1 = nn.Sequential(
                            nn.Conv2d(3, 16, (3, 3)),
                            nn.ReLU(),
                            nn.BatchNorm2d(16),
                            nn.MaxPool2d((2, 2)),
                        )
                        self.conv2 = nn.Sequential(
                            nn.Conv2d(16, 32, (3, 3)),
                            nn.ReLU(),
                            nn.BatchNorm2d(32),
                            nn.MaxPool2d((2, 2)),
                        )
                        self.conv3 = nn.Sequential(
                            nn.Conv2d(32, 64, (3, 3)),
                            nn.ReLU(),
                            nn.BatchNorm2d(64),
                            nn.MaxPool2d((2, 2)),
                        )
                        self.ln1 = nn.Linear(64 * 26 * 26, 16)
                        self.relu = nn.ReLU()
                        self.batchnorm = nn.BatchNorm1d(16)
                        self.dropout = nn.Dropout2d(0.5)
                        self.ln2 = nn.Linear(16, 5)

                        self.ln4 = nn.Linear(13, 10) #tabular input size
                        self.ln5 = nn.Linear(10, 10)
                        self.ln6 = nn.Linear(10, 5)
                        self.ln7 = nn.Linear(10, 1)

                    def forward(self, img, tab):
                        img = self.conv1(img)
                        img = self.conv2(img)
                        img = self.conv3(img)
                        img = img.reshape(img.shape[0], -1)
                        img = self.ln1(img)
                        img = self.relu(img)
                        img = self.batchnorm(img)
                        img = self.dropout(img)
                        img = self.ln2(img)
                        img = self.relu(img)

                        tab = self.ln4(tab)
                        tab = self.relu(tab)
                        tab = self.ln5(tab)
                        tab = self.relu(tab)
                        tab = self.ln6(tab)
                        tab = self.relu(tab)

                        x = torch.cat((img, tab), dim=1)
                        x = self.relu(x)

                        return self.ln7(x)

                return TabImg()

            def forward(self, img, tab):
                return self._model(img, tab)

            def training_step(self, batch, batch_idx):
                x, y = batch
                tabular, image, *support = self.split_dataloader_output(x)
                criterion = torch.nn.L1Loss()

                y_pred = torch.flatten(self(image, tabular))

                y_pred = y_pred.double()

                loss = criterion(y_pred, y)

                return loss

            def validation_step(self, batch, batch_idx):
                x, y = batch
                tabular, image, *support = self.split_dataloader_output(x)
                criterion = torch.nn.L1Loss()
                y_pred = torch.flatten(self(image, tabular))
                y_pred = y_pred.double()
                acc = accuracy(y_pred, y)
                val_loss = criterion(y_pred, y)

                return {
                    "validation_loss": val_loss,
                    "validation_accuracy": acc
                }

            def test_step(self, batch, batch_idx):
                x, y = batch
                tabular, image, *support = self.split_dataloader_output(x)
                criterion = torch.nn.L1Loss()
                y_pred = torch.flatten(self(image, tabular))
                y_pred = y_pred.double()

                # Output targets and prediction for later
                return {"predictions": y_pred, "targets": y}

            def split_dataloader_output(self, data):
                tab, images, sup = data
                weights = sup[:, 0].float()
                if sup.shape[1] > 2:
                    category = sup[:, -1].long()
                else:
                    category = None
                return tab.float(), images, weights, category

            def configure_optimizers(self):
                return torch.optim.Adam(self.parameters(), lr=(self.learning_rate))
        """
    )


@fixture
def pytorch_bitfount_segmentation_model() -> str:
    """Pytorch UNet model for segmentation tasks."""
    # Based on https://github.com/milesial/Pytorch-UNet/tree/master/unet
    return inspect.cleandoc(
        """
        import torch
        import torch.nn as nn
        import torch.nn.functional as F
        import torchmetrics

        from bitfount.backends.pytorch import PyTorchBitfountModel
        from bitfount.backends.pytorch.loss import SoftDiceLoss



        class DummyUnet(PyTorchBitfountModel):
         # Implementation of a UNet model, used for testing purposes.
            def __init__(self, n_channels, n_classes, **kwargs):
                super().__init__(**kwargs)

                self.n_channels = n_channels
                self.n_classes = n_classes
                self.bilinear = True
                self.dice_loss = SoftDiceLoss()
                self.ce_loss = torch.nn.CrossEntropyLoss()


                def double_conv(in_channels, out_channels):
                    return nn.Sequential(
                        nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1),
                        nn.BatchNorm2d(out_channels),
                        nn.ReLU(inplace=True),
                        nn.Conv2d(out_channels, out_channels, kernel_size=3, padding=1),
                        nn.BatchNorm2d(out_channels),
                        nn.ReLU(inplace=True),
                    )

                def down(in_channels, out_channels):
                    return nn.Sequential(
                        nn.MaxPool2d(2),
                        double_conv(in_channels, out_channels)
                    )

                class up(nn.Module):
                    def __init__(self, in_channels, out_channels, bilinear=True):
                        super().__init__()

                        if bilinear:
                            self.up = nn.Upsample(
                            scale_factor=2, mode='bilinear', align_corners=True
                            )
                        else:
                            self.up = nn.ConvTranpose2d(in_channels // 2, in_channels // 2, # noqa: B950
                                                        kernel_size=2, stride=2)

                        self.conv = double_conv(in_channels, out_channels)

                    def forward(self, x1, x2):
                        x1 = self.up(x1)
                        # [Batch size, Channels, Height, Width]
                        diffY = x2.size()[2] - x1.size()[2]
                        diffX = x2.size()[3] - x1.size()[3]

                        x1 = F.pad(x1, [diffX // 2, diffX - diffX // 2,
                                        diffY // 2, diffY - diffY // 2])
                        x = torch.cat([x2, x1], dim=1)
                        return self.conv(x)

                self.inc = double_conv(self.n_channels, 64)
                self.down1 = down(64, 128)
                self.down2 = down(128, 256)
                self.down3 = down(256, 512)
                self.down4 = down(512, 512)
                self.up1 = up(1024, 256)
                self.up2 = up(512, 128)
                self.up3 = up(256, 64)
                self.up4 = up(128, 64)
                self.out = nn.Conv2d(64, self.n_classes, kernel_size=1)
            def create_model(self):
                pass
            def forward(self, x):
                x1 = self.inc(x)
                x2 = self.down1(x1)
                x3 = self.down2(x2)
                x4 = self.down3(x3)
                x5 = self.down4(x4)
                x = self.up1(x5, x4)
                x = self.up2(x, x3)
                x = self.up3(x, x2)
                x = self.up4(x, x1)
                return self.out(x)

            def split_dataloader_output(self, data):
                images, sup = data
                weights = sup[:, 0].float()
                if sup.shape[1] > 2:
                    category = sup[:, -1].long()
                else:
                    category = None
                return images, weights, category

            def training_step(self, batch, batch_nb):
                x, y = batch
                x, *sup = self.split_dataloader_output(x)
                y = y[:, 0].long()
                y_hat = self.forward(x)

                # Cross entropy loss
                ce_loss = F.cross_entropy(y_hat, y) if self.n_classes > 1 else F.binary_cross_entropy_with_logits(y_hat, y)  # noqa: B950

                return {'loss': ce_loss}

            def validation_step(self, batch, batch_nb):
                x, y = batch
                x, *sup = self.split_dataloader_output(x)
                # Get rid of the number of channels dimension and make targets of type `long`
                y = y[:, 0].long()
                y_hat = self.forward(x)
                softmax_y_hat = F.softmax(y_hat, dim=1)

                # Cross entropy loss
                ce_loss = F.cross_entropy(y_hat, y) if self.n_classes > 1 else F.binary_cross_entropy_with_logits(y_hat, y)  # noqa: B950
                # dice loss
                dice_loss = self.dice_loss(softmax_y_hat, y)
                # total loss
                total_loss = (ce_loss + dice_loss) / 2

                # torchmetrics dice score
                dice_pytorch = torchmetrics.functional.dice_score(softmax_y_hat,y)
                # torchmetrics Jaccard index/ IoU (intersection vs union)
                iou = torchmetrics.functional.jaccard_index(y_hat, y, num_classes=3)
                return {
                'ce_loss': ce_loss,
                'dice_loss': dice_loss,
                'loss': total_loss,
                'iou': iou,
                "dice_score": dice_pytorch
                }

            def validation_epoch_end(self, outputs):
                mean_outputs = {}
                for k in outputs[0].keys():
                    mean_outputs[k] = torch.stack([x[k] for x in outputs]).mean()
                # Add the means to the validation stats.
                self.val_stats.append(mean_outputs)

                # Also log out these averaged metrics
                for k, v in mean_outputs.items():
                    self.log(f"avg_{k}", v)

            def test_step(self, batch, batch_nb):
                data, y = batch
                x, *sup = self.split_dataloader_output(data)

                # Get rid of the number of channels dimension and make targets of type `long`
                y = y[:, 0].long()

                # Get validation output and predictions
                y_hat = self.forward(x)
                pred = F.softmax(y_hat, dim=1)

                # Output targets and prediction for later
                return {"predictions": pred, "targets": y}


            def configure_optimizers(self):
                return torch.optim.Adam(self.parameters(), lr=1e-4)
        """
    )



"""Pytest configuration for protocols tests."""
from unittest.mock import Mock, create_autospec

from pytest import fixture

from bitfount.federated.transport.modeller_transport import _ModellerMailbox
from bitfount.federated.transport.worker_transport import _WorkerMailbox
from bitfount.hub.api import BitfountHub
from bitfount.hub.authentication_flow import BitfountSession


@fixture
def mock_algorithm() -> Mock:
    """Returns mock untyped algorithm."""
    return Mock()


@fixture
def mock_modeller_mailbox() -> Mock:
    """Returns mock mailbox."""
    mock_modeller_mailbox: Mock = create_autospec(_ModellerMailbox, instance=True)
    return mock_modeller_mailbox


@fixture
def mock_worker_mailbox() -> Mock:
    """Returns mock mailbox."""
    mock_worker_mailbox: Mock = create_autospec(_WorkerMailbox, instance=True)
    return mock_worker_mailbox


@fixture
def mock_hub() -> Mock:
    """Returns mock BitfountHub."""
    mock_hub: Mock = create_autospec(BitfountHub, instance=True)
    mock_hub.session = create_autospec(BitfountSession, instance=True)
    return mock_hub



"""Tests that modeller can initiate training with Pods."""
from __future__ import annotations

from collections import namedtuple
from datetime import datetime, timedelta, timezone
from typing import Callable, Dict, Iterable, List, Mapping, Optional, Tuple
from unittest.mock import Mock, call, create_autospec

from _pytest.logging import LogCaptureFixture
from grpc import RpcError
import msgpack
from pytest import fixture, raises
from pytest_mock import MockerFixture

from bitfount.federated.authorisation_checkers import IdentityVerificationMethod
from bitfount.federated.encryption import _AESEncryption, _RSAEncryption
from bitfount.federated.exceptions import BitfountTaskStartError
from bitfount.federated.task_requests import (
    _EncryptedTaskRequest,
    _SignedEncryptedTaskRequest,
    _TaskRequest,
    _TaskRequestMessage,
)
from bitfount.federated.transport.base_transport import _run_func_and_listen_to_mailbox
from bitfount.federated.transport.handlers import _AsyncMultipleResponsesHandler
from bitfount.federated.transport.message_service import (
    _BitfountMessage,
    _BitfountMessageType,
    _DecryptedBitfountMessage,
    _MessageEncryption,
    _MessageService,
)
from bitfount.federated.transport.modeller_transport import (
    _get_parameter_updates_from_workers,
    _get_psi_datasets_from_workers,
    _get_public_key,
    _get_training_metrics_from_workers,
    _ModellerMailbox,
    _parameter_updates_handler,
    _psi_dataset_handler,
    _public_key_handler,
    _training_metrics_handler,
    _WorkerMailboxDetails,
)
from bitfount.federated.transport.types import (
    _OIDCAuthFlowResponse,
    _OIDCClientID,
    _PodDeviceCodeDetails,
)
from bitfount.federated.types import (
    SerializedAlgorithm,
    SerializedProtocol,
    _PodResponseType,
    _TaskRequestMessageGenerator,
)
from bitfount.types import _SerializedWeights
from tests.utils import PytestRequest
from tests.utils.fixtures import (
    KeyBasedGenerator,
    MessageOrException,
    PollForMessagesPatcher,
)
from tests.utils.helper import unit_test

WorkerDetails = namedtuple(
    "WorkerDetails", ["pod_identifier", "mailbox_id", "aes_encryption_key"]
)


@fixture
def modeller_name() -> str:
    """Name of the modeller."""
    return "modeller"


@fixture
def modeller_mailbox_id() -> str:
    """A fake modeller mailbox id."""
    return "modeller_mailbox_id"


@fixture
def pod_identifiers() -> List[str]:
    """List of pod identifiers."""
    return ["user1/pod_1", "user2/pod_2", "user3/pod_3"]


@fixture
def worker_mailbox_ids(pod_identifiers: List[str]) -> Dict[str, str]:
    """A dict of pod identifiers to worker mailboxes IDs."""
    return {
        pod_identifier: f"worker_mailbox_id_{idx}"
        for idx, pod_identifier in enumerate(pod_identifiers)
    }


@fixture
def worker_mailbox_details(worker_mailbox_ids: Dict[str, str]) -> List[WorkerDetails]:
    """Provides an indexable version of worker mailbox IDs."""
    # Need to be able to more easily index into the worker_mailbox_ids
    worker_details = [
        WorkerDetails(pod_identifier, mailbox_id, b"aes_encryption_key")
        for pod_identifier, mailbox_id in worker_mailbox_ids.items()
    ]
    return worker_details


@fixture
def pod_public_keys(pod_identifiers: List[str]) -> Dict[str, Mock]:
    """Mocked out public keys for pods."""
    return {pod_identifier: Mock() for pod_identifier in pod_identifiers}


@fixture
def aes_keys() -> List[bytes]:
    """AES keys to use in encrypting."""
    return [_AESEncryption.generate_key() for _ in range(3)]


@fixture
def worker_mailboxes(
    aes_keys: List[bytes],
    pod_public_keys: Dict[str, Mock],
    worker_mailbox_ids: Dict[str, str],
) -> Dict[str, _WorkerMailboxDetails]:
    """A set of worker mailboxes to use."""
    return {
        pod_identifier: _WorkerMailboxDetails(
            pod_identifier=pod_identifier,
            public_key=pod_public_keys[pod_identifier],
            mailbox_id=mailbox_id,
            aes_encryption_key=aes_keys[idx],
        )
        for idx, (pod_identifier, mailbox_id) in enumerate(worker_mailbox_ids.items())
    }


@fixture
def mock_message_service() -> Mock:
    """A mocked MessageService."""
    mock_message_service: Mock = create_autospec(_MessageService, instance=True)
    return mock_message_service


@fixture
def idp_url() -> str:
    """Identity Provider URL."""
    return "https://idp-url.unit-testing.bitfount.com"


@fixture
def task_id() -> str:
    """Task ID."""
    return "this-is-a-task-id"


@fixture(params=(True, False), ids=("task_id_incl", "no_task_id"))
def opt_task_id(request: PytestRequest, task_id: str) -> Optional[str]:
    """Returns a task ID or None, to cover both cases."""
    incl_task_id: bool = request.param
    if incl_task_id:
        return task_id
    else:
        return None


@fixture
def online_check_uuid() -> str:
    """UUID for online checking."""
    return "auuidforonlinecheck"


@fixture
def modeller_mailbox(
    idp_url: str,
    mock_message_service: Mock,
    modeller_mailbox_id: str,
    task_id: str,
    worker_mailboxes: Dict[str, _WorkerMailboxDetails],
) -> _ModellerMailbox:
    """The ModellerMailbox under test."""
    return _ModellerMailbox(
        mailbox_id="modeller_mailbox_id",
        worker_mailboxes=worker_mailboxes,
        message_service=mock_message_service,
        task_id=task_id,
    )


@fixture
def signatures() -> List[bytes]:
    """List of fake signatures for message signing."""
    signatures = [b"signature1", b"signature2", b"signature3"]
    return signatures


@fixture
def mock_signer(mocker: MockerFixture, signatures: List[bytes]) -> Mock:
    """Mock RSA signing with predefined signatures to return."""
    mock_signer = mocker.patch.object(_RSAEncryption, "sign_message")
    mock_signer.side_effect = signatures
    return mock_signer


@fixture
def encrypted_task_requests() -> List[bytes]:
    """List of fake encrypted task requests."""
    encrypted_task_requests = [b"first", b"second", b"third"]
    return encrypted_task_requests


@fixture
def mock_rsa_encrypter(
    encrypted_task_requests: List[bytes], mocker: MockerFixture
) -> Mock:
    """Mock RSA encryption with predefined encrypted messages to return."""
    mock_rsa_encrypter = mocker.patch.object(_RSAEncryption, "encrypt")
    mock_rsa_encrypter.side_effect = encrypted_task_requests
    return mock_rsa_encrypter


@fixture
def mock_aes_decrypter(mocker: MockerFixture) -> Mock:
    """Mock AES decryption for cases where message bodies aren't encrypted."""
    # If the message bodies are not encrypted we need to mock out the decryption
    # and instead replace it with just returning message bodies.
    mock_decrypter = mocker.patch.object(_MessageEncryption, "decrypt_incoming_message")
    mock_decrypter.side_effect = lambda body, key: body
    return mock_decrypter


@fixture
def serialized_protocol() -> SerializedProtocol:
    """Protocol details."""
    return SerializedProtocol(
        class_name="some protocol",
        algorithm=SerializedAlgorithm(class_name="some algorithm"),
    )


@unit_test
class TestModellerMailbox:
    """Tests for ModellerMailbox class."""

    async def test_send_task_requests(
        self,
        default_task_request_msg_gen: _TaskRequestMessageGenerator,
        idp_url: str,
        mock_message_service: Mock,
        mocker: MockerFixture,
        modeller_mailbox_id: str,
        task_id: str,
        worker_mailboxes: Dict[str, _WorkerMailboxDetails],
    ) -> None:
        """Test send_task_requests factory method.

        This is the class method and so should return us a created ModellerMailbox.
        """
        # Mock out underlying send method
        mocked_sender = mocker.patch.object(_ModellerMailbox, "_send_task_requests")
        mocked_sender.return_value = (modeller_mailbox_id, worker_mailboxes, task_id)

        modeller_mailbox = await _ModellerMailbox.send_task_requests(
            serialized_protocol=SerializedProtocol(
                class_name="some protocol",
                algorithm=SerializedAlgorithm(class_name="some algorithm"),
            ),
            pod_public_keys=Mock(),
            task_request_msg_gen=default_task_request_msg_gen,
            message_service=mock_message_service,
        )

        assert modeller_mailbox.mailbox_id == modeller_mailbox_id
        assert modeller_mailbox.worker_mailboxes == worker_mailboxes
        assert modeller_mailbox.message_service == mock_message_service
        assert modeller_mailbox._task_id == task_id

    async def test__send_task_requests_is_successful(
        self,
        aes_keys: List[bytes],
        encrypted_task_requests: List[bytes],
        key_based_task_request_msg_gen: KeyBasedGenerator,
        mock_message_service: Mock,
        mock_rsa_encrypter: Mock,
        mock_signer: Mock,
        mocker: MockerFixture,
        modeller_mailbox_id: str,
        pod_identifiers: List[str],
        pod_public_keys: Dict[str, Mock],
        serialized_protocol: SerializedProtocol,
        signatures: List[bytes],
        task_id: str,
        worker_mailbox_ids: Dict[str, str],
        worker_mailboxes: Dict[str, _WorkerMailboxDetails],
    ) -> None:
        """Happy path, job requests are all sent successfully.

        Key-based identity verification.
        """
        # Mock AES key generation
        mock_key_generator = mocker.patch.object(_AESEncryption, "generate_key")
        mock_key_generator.side_effect = aes_keys

        # Mock the message services setup method
        mock_message_service.setup_communication_with_pods.return_value = (
            modeller_mailbox_id,
            worker_mailbox_ids,
            task_id,
        )

        # Mock/create other args
        private_key = key_based_task_request_msg_gen.key
        task_request_msg_gen = key_based_task_request_msg_gen.gen

        # Call method
        (
            received_modeller_mailbox_id,
            received_worker_mailboxes,
            received_task_id,
        ) = await _ModellerMailbox._send_task_requests(
            serialized_protocol=serialized_protocol,
            pod_public_keys=pod_public_keys,
            task_request_msg_gen=task_request_msg_gen,
            message_service=mock_message_service,
        )

        # Check return values
        assert received_modeller_mailbox_id == modeller_mailbox_id
        assert received_worker_mailboxes == worker_mailboxes
        assert received_task_id == task_id

        # Check message service method call
        mock_message_service.setup_communication_with_pods.assert_called_once_with(
            {
                pod_identifier: _TaskRequestMessage(
                    serialized_protocol=serialized_protocol,
                    auth_type=IdentityVerificationMethod.KEYS.value,
                    request=_SignedEncryptedTaskRequest(
                        encrypted_request=encrypted_task_requests[i],
                        signature=signatures[i],
                    ).serialize(),
                ).serialize()
                for i, pod_identifier in enumerate(pod_identifiers)
            },
        )

        # Ensure we encrypt and sign correctly
        mock_rsa_encrypter.assert_has_calls(
            [
                # Each call to the encryption should be the (task request body,
                # the list of all pods involved, generated AES key) and the public
                # key to encrypt with.
                call(
                    _TaskRequest(
                        serialized_protocol=serialized_protocol,
                        pod_identifiers=pod_identifiers,
                        aes_key=aes_keys[i],
                    ).serialize(),
                    pod_public_key,
                )
                for i, pod_public_key in enumerate(pod_public_keys.values())
            ]
        )
        mock_signer.assert_has_calls(
            # Each call to the signer should be the private key to sign with and
            # the encrypted message body (i.e. task).
            [call(private_key, etr) for etr in encrypted_task_requests]
        )

    async def test__send_task_requests_is_successful_when_some_pods_are_offline(
        self,
        aes_keys: List[bytes],
        encrypted_task_requests: List[bytes],
        key_based_task_request_msg_gen: KeyBasedGenerator,
        mock_message_service: Mock,
        mock_rsa_encrypter: Mock,
        mock_signer: Mock,
        mocker: MockerFixture,
        modeller_mailbox_id: str,
        pod_identifiers: List[str],
        pod_public_keys: Dict[str, Mock],
        serialized_protocol: SerializedProtocol,
        signatures: List[bytes],
        task_id: str,
        worker_mailbox_ids: Dict[str, str],
        worker_mailboxes: Dict[str, _WorkerMailboxDetails],
    ) -> None:
        """Some pods are offline, so there aren't mailboxes returned for them.

        In the case where some pods are offline we continue the task,
        except we only receive mailboxes for some of the pods for the
        message service.

        We then continue training, but only with the online tasks.
        """
        # Mock AES key generation
        mock_key_generator = mocker.patch.object(_AESEncryption, "generate_key")
        mock_key_generator.side_effect = aes_keys

        # For this test we ensure the message service returns fewer
        # pods than were requested.
        worker_mailboxes_from_message_service = dict(worker_mailbox_ids)
        worker_mailboxes_from_message_service.popitem()

        # Mock the message services setup method
        mock_message_service.setup_communication_with_pods.return_value = (
            modeller_mailbox_id,
            worker_mailboxes_from_message_service,
            task_id,
        )

        # Mock/create other args
        private_key = key_based_task_request_msg_gen.key
        task_request_msg_gen = key_based_task_request_msg_gen.gen

        # Call method
        (
            received_modeller_mailbox_id,
            received_worker_mailboxes,
            received_task_id,
        ) = await _ModellerMailbox._send_task_requests(
            serialized_protocol=serialized_protocol,
            pod_public_keys=pod_public_keys,
            task_request_msg_gen=task_request_msg_gen,
            message_service=mock_message_service,
        )

        # Modeller mailbox ID matches the mocked message service response
        assert received_modeller_mailbox_id == modeller_mailbox_id
        # We only create WorkerMailboxDetails for the mailboxes
        # returned by the message service
        worker_mailboxes.popitem()
        assert received_worker_mailboxes == worker_mailboxes
        assert received_task_id == task_id

        # Check message service method call included all pod identifiers
        # This includes the ones that aren't online, we don't know that one is
        # offline until the message service tells us!
        mock_message_service.setup_communication_with_pods.assert_called_once_with(
            {
                pod_identifier: _TaskRequestMessage(
                    serialized_protocol=serialized_protocol,
                    auth_type=IdentityVerificationMethod.KEYS.value,
                    request=_SignedEncryptedTaskRequest(
                        encrypted_request=encrypted_task_requests[i],
                        signature=signatures[i],
                    ).serialize(),
                ).serialize()
                for i, pod_identifier in enumerate(worker_mailbox_ids.keys())
            },
        )

        # Ensure we encrypt and sign correctly the first task messages
        # This includes the ones that aren't online, we don't know that one is
        # offline until the message service tells us!
        mock_rsa_encrypter.assert_has_calls(
            [
                # Each call to the encryption should be the (task request body,
                # the list of all pods involved, generated AES key) and the public
                # key to encrypt with.
                call(
                    _TaskRequest(
                        serialized_protocol=serialized_protocol,
                        pod_identifiers=pod_identifiers,
                        aes_key=aes_keys[i],
                    ).serialize(),
                    pod_public_key,
                )
                for i, pod_public_key in enumerate(pod_public_keys.values())
            ]
        )
        mock_signer.assert_has_calls(
            # Each call to the signer should be the private key to sign with and
            # the encrypted message body (i.e. task).
            [call(private_key, etr) for etr in encrypted_task_requests]
        )

    async def test__send_task_requests_fails(
        self,
        encrypted_task_requests: List[bytes],
        key_based_task_request_msg_gen: KeyBasedGenerator,
        mock_message_service: Mock,
        mock_rsa_encrypter: Mock,
        mock_signer: Mock,
        pod_identifiers: List[str],
        pod_public_keys: Dict[str, Mock],
        rpc_error: RpcError,
        serialized_protocol: SerializedProtocol,
        signatures: List[bytes],
    ) -> None:
        """RPC Error occurs for one attempt at starting training.

        Uses key-based identity verification method.
        """
        # Mock out error raising in message service.
        mock_message_service.setup_communication_with_pods.side_effect = rpc_error

        # Mock/create other args
        task_request_msg_gen = key_based_task_request_msg_gen.gen

        # Call method
        with raises(RuntimeError, match="Failed to start task with pods:"):
            await _ModellerMailbox._send_task_requests(
                serialized_protocol=serialized_protocol,
                pod_public_keys=pod_public_keys,
                task_request_msg_gen=task_request_msg_gen,
                message_service=mock_message_service,
            )

        # Check that the encrypted and signed messages were passed to
        # setup_communication_with_pods
        mock_message_service.setup_communication_with_pods.assert_called_once_with(
            {
                pod_identifier: _TaskRequestMessage(
                    serialized_protocol=serialized_protocol,
                    auth_type=IdentityVerificationMethod.KEYS.value,
                    request=_SignedEncryptedTaskRequest(
                        encrypted_request=encrypted_task_requests[i],
                        signature=signatures[i],
                    ).serialize(),
                ).serialize()
                for i, pod_identifier in enumerate(pod_identifiers)
            },
        )

    async def test__send_task_requests_has_no_signatures_if_saml_based(
        self,
        aes_keys: List[bytes],
        encrypted_task_requests: List[bytes],
        mock_message_service: Mock,
        mock_rsa_encrypter: Mock,
        mocker: MockerFixture,
        modeller_mailbox_id: str,
        pod_identifiers: List[str],
        pod_public_keys: Dict[str, Mock],
        saml_task_request_msg_gen: _TaskRequestMessageGenerator,
        serialized_protocol: SerializedProtocol,
        task_id: str,
        worker_mailbox_ids: Dict[str, str],
        worker_mailboxes: Dict[str, _WorkerMailboxDetails],
    ) -> None:
        """Tests _send_task_requests has no sigs if SAML-based verification."""
        # Mock AES key generation
        mock_key_generator = mocker.patch.object(_AESEncryption, "generate_key")
        mock_key_generator.side_effect = aes_keys

        # Mock the message services setup method
        mock_message_service.setup_communication_with_pods.return_value = (
            modeller_mailbox_id,
            worker_mailbox_ids,
            task_id,
        )

        # Call method
        (
            received_modeller_mailbox_id,
            received_worker_mailboxes,
            received_task_id,
        ) = await _ModellerMailbox._send_task_requests(
            serialized_protocol=serialized_protocol,
            pod_public_keys=pod_public_keys,
            task_request_msg_gen=saml_task_request_msg_gen,
            message_service=mock_message_service,
        )

        # Check return values
        assert received_modeller_mailbox_id == modeller_mailbox_id
        assert received_worker_mailboxes == worker_mailboxes
        assert received_task_id == task_id

        # Check message service method call
        mock_message_service.setup_communication_with_pods.assert_called_once_with(
            {
                pod_identifier: _TaskRequestMessage(
                    serialized_protocol=serialized_protocol,
                    auth_type=IdentityVerificationMethod.SAML.value,
                    request=_EncryptedTaskRequest(
                        encrypted_request=encrypted_task_requests[i],
                    ).serialize(),
                ).serialize()
                for i, pod_identifier in enumerate(pod_identifiers)
            },
        )

        # Ensure we encrypt correctly
        mock_rsa_encrypter.assert_has_calls(
            [
                # Each call to the encryption should be the (task request body,
                # the list of all pods involved, generated AES key) and the public
                # key to encrypt with.
                call(
                    _TaskRequest(
                        serialized_protocol=serialized_protocol,
                        pod_identifiers=pod_identifiers,
                        aes_key=aes_keys[i],
                    ).serialize(),
                    pod_public_key,
                )
                for i, pod_public_key in enumerate(pod_public_keys.values())
            ]
        )

    async def test_process_task_request_responses(
        self,
        mock_aes_decrypter: Mock,
        mock_message_service: Mock,
        modeller_mailbox: _ModellerMailbox,
        modeller_mailbox_id: str,
        modeller_name: str,
        opt_task_id: Optional[str],
        patch_poll_for_messages: PollForMessagesPatcher,
        worker_mailbox_details: List[WorkerDetails],
        worker_mailbox_ids: Dict[str, str],
        worker_mailboxes: Dict[str, _WorkerMailboxDetails],
    ) -> None:
        """Happy path, all pods send responses to training requests."""
        # Create fake task response messages
        task_response_messages = [
            _BitfountMessage(
                message_type=_BitfountMessageType.JOB_ACCEPT,
                body=msgpack.dumps(
                    {
                        _PodResponseType.ACCEPT.name: worker_mailbox_details[
                            0
                        ].pod_identifier
                    }
                ),
                recipient=modeller_name,
                recipient_mailbox_id=modeller_mailbox_id,
                sender=worker_mailbox_details[0].pod_identifier,
                sender_mailbox_id=worker_mailbox_details[0].mailbox_id,
                task_id=opt_task_id,
            ),
            _BitfountMessage(
                message_type=_BitfountMessageType.JOB_ACCEPT,
                body=msgpack.dumps(
                    {
                        _PodResponseType.ACCEPT.name: worker_mailbox_details[
                            1
                        ].pod_identifier
                    }
                ),
                recipient=modeller_name,
                recipient_mailbox_id=modeller_mailbox_id,
                sender=worker_mailbox_details[1].pod_identifier,
                sender_mailbox_id=worker_mailbox_details[1].mailbox_id,
                task_id=opt_task_id,
            ),
            _BitfountMessage(
                message_type=_BitfountMessageType.JOB_REJECT,
                body=msgpack.dumps(
                    {
                        _PodResponseType.NO_ACCESS.name: "some/pod",
                        _PodResponseType.NO_ACCESS.name: "some/pod",
                    }
                ),
                recipient=modeller_name,
                recipient_mailbox_id=modeller_mailbox_id,
                sender=worker_mailbox_details[2].pod_identifier,
                sender_mailbox_id=worker_mailbox_details[2].mailbox_id,
                task_id=opt_task_id,
            ),
        ]

        # Assign these to poll_for_messages which is the underlying method we
        # expect to yield messages.
        patch_poll_for_messages(task_response_messages)

        # Start listening and processing messages
        accepted_pod_mailboxes = await _run_func_and_listen_to_mailbox(
            modeller_mailbox.process_task_request_responses(), modeller_mailbox
        )

        # Pod3 rejected the task, so we aren't expecting it!
        worker_mailboxes.pop(worker_mailbox_details[2].pod_identifier)
        assert accepted_pod_mailboxes == worker_mailboxes

    async def test_get_saml_challenges_retrieves_saml_responses(
        self,
        mock_aes_decrypter: Mock,
        modeller_mailbox: _ModellerMailbox,
        modeller_mailbox_id: str,
        modeller_name: str,
        opt_task_id: Optional[str],
        patch_poll_for_messages: PollForMessagesPatcher,
        worker_mailbox_details: List[WorkerDetails],
    ) -> None:
        """Happy path, all pods send responses to training requests."""
        saml_challenges = [
            msgpack.dumps(bytes(f"saml challenge {i}", "utf-8")) for i in range(3)
        ]

        # Create fake task response messages
        saml_requests = [
            _BitfountMessage(
                message_type=_BitfountMessageType.SAML_REQUEST,
                body=saml_challenges[0],
                recipient=modeller_name,
                recipient_mailbox_id=modeller_mailbox_id,
                sender=worker_mailbox_details[0].pod_identifier,
                sender_mailbox_id=worker_mailbox_details[0].mailbox_id,
                task_id=opt_task_id,
            ),
            _BitfountMessage(
                message_type=_BitfountMessageType.SAML_REQUEST,
                body=saml_challenges[1],
                recipient=modeller_name,
                recipient_mailbox_id=modeller_mailbox_id,
                sender=worker_mailbox_details[1].pod_identifier,
                sender_mailbox_id=worker_mailbox_details[1].mailbox_id,
                task_id=opt_task_id,
            ),
            _BitfountMessage(
                message_type=_BitfountMessageType.SAML_REQUEST,
                body=saml_challenges[2],
                recipient=modeller_name,
                recipient_mailbox_id=modeller_mailbox_id,
                sender=worker_mailbox_details[2].pod_identifier,
                sender_mailbox_id=worker_mailbox_details[2].mailbox_id,
                task_id=opt_task_id,
            ),
        ]

        # Assign these to poll_for_messages which is the underlying method we
        # expect to yield messages.
        patch_poll_for_messages(saml_requests)

        # Start listening and processing messages
        received_saml_challenges = await _run_func_and_listen_to_mailbox(
            modeller_mailbox.get_saml_challenges(), modeller_mailbox
        )

        # Ensure messages returned match those that we mocked
        assert received_saml_challenges == [
            challenge.decrypt(
                modeller_mailbox.worker_mailboxes[challenge.sender].aes_encryption_key
            )
            for challenge in saml_requests
        ]

    async def test_get_saml_challenges_throws_error(
        self,
        modeller_mailbox: _ModellerMailbox,
        patch_poll_for_messages: PollForMessagesPatcher,
        rpc_error: RpcError,
    ) -> None:
        """Test Start Task error thrown on error."""
        patch_poll_for_messages([rpc_error])
        with raises(
            BitfountTaskStartError, match="Failed to start task with all pods."
        ):
            await _run_func_and_listen_to_mailbox(
                modeller_mailbox.get_saml_challenges(),
                modeller_mailbox,
            )

    async def test_handles_pod_never_responds(
        self,
        mock_aes_decrypter: Mock,
        mock_message_service: Mock,
        modeller_mailbox: _ModellerMailbox,
        modeller_mailbox_id: str,
        modeller_name: str,
        opt_task_id: Optional[str],
        patch_poll_for_messages: PollForMessagesPatcher,
        worker_mailbox_details: List[WorkerDetails],
        worker_mailboxes: Dict[str, _WorkerMailboxDetails],
    ) -> None:
        """Some pods never send a response to the training request."""
        # Create only a single response message
        task_response_messages = [
            _BitfountMessage(
                message_type=_BitfountMessageType.JOB_ACCEPT,
                body=msgpack.dumps(
                    {
                        _PodResponseType.ACCEPT.name: worker_mailbox_details[
                            0
                        ].pod_identifier
                    }
                ),
                recipient=modeller_name,
                recipient_mailbox_id=modeller_mailbox_id,
                sender=worker_mailbox_details[0].pod_identifier,
                sender_mailbox_id=worker_mailbox_details[0].mailbox_id,
                task_id=opt_task_id,
            ),
        ]

        # Assign this message to poll_for_messages yield
        patch_poll_for_messages(task_response_messages)

        # Start listening and processing messages. We timeout on the responses to
        # simulate non-responding pods. This may need to be increased if the test
        # is failing but should need to be no higher than 5 seconds.
        accepted_pod_mailboxes = await _run_func_and_listen_to_mailbox(
            modeller_mailbox.process_task_request_responses(timeout=1), modeller_mailbox
        )

        # Remove non-responding pods from expected results
        worker_mailboxes.pop(worker_mailbox_details[1].pod_identifier)
        worker_mailboxes.pop(worker_mailbox_details[2].pod_identifier)

        assert accepted_pod_mailboxes == worker_mailboxes

    async def test_handles_message_retrieval_exception(
        self,
        mock_aes_decrypter: Mock,
        mock_message_service: Mock,
        modeller_mailbox: _ModellerMailbox,
        modeller_mailbox_id: str,
        modeller_name: str,
        opt_task_id: Optional[str],
        patch_poll_for_messages: PollForMessagesPatcher,
        rpc_error: RpcError,
        worker_mailbox_details: List[WorkerDetails],
    ) -> None:
        """RPC Exception when trying to check for responses from pods."""
        task_response_messages: List[MessageOrException] = [
            _BitfountMessage(
                message_type=_BitfountMessageType.JOB_ACCEPT,
                body=msgpack.dumps(
                    {
                        _PodResponseType.ACCEPT.name: worker_mailbox_details[
                            0
                        ].pod_identifier
                    }
                ),
                recipient=modeller_name,
                recipient_mailbox_id=modeller_mailbox_id,
                sender=worker_mailbox_details[0].pod_identifier,
                sender_mailbox_id=worker_mailbox_details[0].mailbox_id,
                task_id=opt_task_id,
            ),
            rpc_error,
        ]

        patch_poll_for_messages(task_response_messages)

        with raises(
            BitfountTaskStartError, match="Failed to start task with all pods."
        ):
            await _run_func_and_listen_to_mailbox(
                modeller_mailbox.process_task_request_responses(),
                modeller_mailbox,
            )

    async def test_send_to_all_pods_aes_encrypt_successful(
        self,
        mock_message_service: Mock,
        mock_message_timestamps: Callable[[Iterable[str]], Mock],
        mocker: MockerFixture,
        modeller_mailbox: _ModellerMailbox,
        modeller_mailbox_id: str,
        modeller_name: str,
        pod_identifiers: List[str],
        task_id: str,
        worker_mailboxes: Dict[str, _WorkerMailboxDetails],
    ) -> None:
        """Message is sent AES encrypted using provided key."""
        fake_timestamps = ["Hello", "World", "!"]
        mock_message_timestamps(fake_timestamps)

        encrypt = mocker.patch.object(_MessageEncryption, "encrypt_outgoing_message")
        encrypted_messages = [
            b"first encrypted",
            b"second encrypted",
            b"third encrypted",
        ]
        encrypt.side_effect = encrypted_messages
        message = "some message"
        dumped_message = msgpack.dumps(message)

        mock_message_service.username = modeller_name

        modeller_mailbox.accepted_worker_mailboxes = worker_mailboxes

        await modeller_mailbox._send_to_all_pods_aes_encrypt(
            message, _BitfountMessageType.TRAINING_COMPLETE
        )

        encrypt.assert_has_calls(
            [
                call(
                    dumped_message,
                    worker_mailboxes[pod_identifiers[0]].aes_encryption_key,
                ),
                call(
                    dumped_message,
                    worker_mailboxes[pod_identifiers[1]].aes_encryption_key,
                ),
                call(
                    dumped_message,
                    worker_mailboxes[pod_identifiers[2]].aes_encryption_key,
                ),
            ]
        )
        assert encrypt.call_count == 3
        mock_message_service.send_message.assert_has_calls(
            [
                call(
                    _BitfountMessage(
                        message_type=_BitfountMessageType.TRAINING_COMPLETE,
                        body=encrypted_messages[0],
                        recipient=pod_identifiers[0],
                        recipient_mailbox_id=worker_mailboxes[
                            pod_identifiers[0]
                        ].mailbox_id,
                        sender=modeller_name,
                        sender_mailbox_id=modeller_mailbox_id,
                        timestamp=fake_timestamps[0],
                        task_id=task_id,
                    ),
                    already_packed=True,
                ),
                call(
                    _BitfountMessage(
                        message_type=_BitfountMessageType.TRAINING_COMPLETE,
                        body=encrypted_messages[1],
                        recipient=pod_identifiers[1],
                        recipient_mailbox_id=worker_mailboxes[
                            pod_identifiers[1]
                        ].mailbox_id,
                        sender=modeller_name,
                        sender_mailbox_id=modeller_mailbox_id,
                        timestamp=fake_timestamps[1],
                        task_id=task_id,
                    ),
                    already_packed=True,
                ),
                call(
                    _BitfountMessage(
                        message_type=_BitfountMessageType.TRAINING_COMPLETE,
                        body=encrypted_messages[2],
                        recipient=pod_identifiers[2],
                        recipient_mailbox_id=worker_mailboxes[
                            pod_identifiers[2]
                        ].mailbox_id,
                        sender=modeller_name,
                        sender_mailbox_id=modeller_mailbox_id,
                        timestamp=fake_timestamps[2],
                        task_id=task_id,
                    ),
                    already_packed=True,
                ),
            ]
        )
        assert mock_message_service.send_message.call_count == 3

    async def test_send_to_all_pods_aes_encrypt_throws_error(
        self,
        mock_message_service: Mock,
        mock_message_timestamps: Callable[[Iterable[str]], Mock],
        mocker: MockerFixture,
        modeller_mailbox: _ModellerMailbox,
        modeller_mailbox_id: str,
        modeller_name: str,
        pod_identifiers: List[str],
        rpc_error: RpcError,
        task_id: str,
        worker_mailboxes: Dict[str, _WorkerMailboxDetails],
    ) -> None:
        """Sending encrypted message receives RpcError."""
        fake_timestamps = ["Hello"]
        mock_message_timestamps(fake_timestamps)

        expected_encrypted_message = b"encrypted_message"
        encrypt = mocker.patch.object(_MessageEncryption, "encrypt_outgoing_message")
        encrypt.return_value = expected_encrypted_message
        message = "some message"
        dumped_message = msgpack.dumps(message)
        mock_message_service.username = modeller_name
        mock_message_service.send_message.side_effect = rpc_error

        modeller_mailbox.accepted_worker_mailboxes = worker_mailboxes

        with raises(RpcError):
            await modeller_mailbox._send_to_all_pods_aes_encrypt(
                message, _BitfountMessageType.ALGORITHM_EXCHANGE
            )

        encrypt.assert_called_once_with(
            dumped_message, worker_mailboxes[pod_identifiers[0]].aes_encryption_key
        )
        mock_message_service.send_message.assert_called_once_with(
            _BitfountMessage(
                message_type=_BitfountMessageType.ALGORITHM_EXCHANGE,
                body=expected_encrypted_message,
                recipient=pod_identifiers[0],
                recipient_mailbox_id=worker_mailboxes[pod_identifiers[0]].mailbox_id,
                sender=modeller_name,
                sender_mailbox_id=modeller_mailbox_id,
                timestamp=fake_timestamps[0],
                task_id=task_id,
            ),
            already_packed=True,
        )

    def test_receive_aes_decrypt_successful(
        self,
        modeller_mailbox: _ModellerMailbox,
        modeller_mailbox_id: str,
        opt_task_id: Optional[str],
        pod_identifiers: List[str],
        worker_mailboxes: Dict[str, _WorkerMailboxDetails],
    ) -> None:
        """Receives and decrypts messages using provided AES key."""
        expected_decrypted_message = b"some message"
        original_message = _BitfountMessage(
            message_type=_BitfountMessageType.TRAINING_COMPLETE,
            body=_MessageEncryption.encrypt_outgoing_message(
                msgpack.dumps(expected_decrypted_message),
                worker_mailboxes[pod_identifiers[1]].aes_encryption_key,
            ),
            recipient=pod_identifiers[0],
            recipient_mailbox_id=worker_mailboxes[pod_identifiers[0]].mailbox_id,
            sender=pod_identifiers[1],
            sender_mailbox_id=modeller_mailbox_id,
            task_id=opt_task_id,
        )
        modeller_mailbox.accepted_worker_mailboxes = worker_mailboxes

        message = modeller_mailbox._decrypt_message(original_message)

        assert message == _DecryptedBitfountMessage(
            message_type=_BitfountMessageType.TRAINING_COMPLETE,
            body=expected_decrypted_message,
            recipient=pod_identifiers[0],
            recipient_mailbox_id=worker_mailboxes[pod_identifiers[0]].mailbox_id,
            sender=pod_identifiers[1],
            sender_mailbox_id=modeller_mailbox_id,
            timestamp=original_message.timestamp,
            task_id=original_message.task_id,
        )

    async def test_send_task_start_message(
        self,
        mocker: MockerFixture,
        modeller_mailbox: _ModellerMailbox,
    ) -> None:
        """Tests modeller sends task start empty message."""
        mock_message_decrypt = mocker.patch.object(
            modeller_mailbox,
            "_send_to_all_pods_aes_encrypt",
            return_value=None,
        )

        await modeller_mailbox.send_task_start_message()
        mock_message_decrypt.assert_awaited_once_with(
            None, _BitfountMessageType.TASK_START
        )

    async def test_send_task_complete_message(
        self,
        mock_message_service: Mock,
        mock_message_timestamps: Callable[[Iterable[str]], Mock],
        mocker: MockerFixture,
        modeller_mailbox: _ModellerMailbox,
        modeller_mailbox_id: str,
        modeller_name: str,
        pod_identifiers: List[str],
        task_id: str,
        worker_mailboxes: Dict[str, _WorkerMailboxDetails],
    ) -> None:
        """Task complete message is sent AES encrypted using provided key."""
        fake_timestamps = ["Hello", "World", "!"]
        mock_message_timestamps(fake_timestamps)
        encrypt = mocker.patch.object(_MessageEncryption, "encrypt_outgoing_message")
        encrypt.side_effect = [None, None, None]
        dumped_message = msgpack.dumps(None)
        mock_message_service.username = modeller_name
        modeller_mailbox.accepted_worker_mailboxes = worker_mailboxes

        await modeller_mailbox.send_task_complete_message()

        encrypt.assert_has_calls(
            [
                call(
                    dumped_message,
                    worker_mailboxes[pod_identifiers[0]].aes_encryption_key,
                ),
                call(
                    dumped_message,
                    worker_mailboxes[pod_identifiers[1]].aes_encryption_key,
                ),
                call(
                    dumped_message,
                    worker_mailboxes[pod_identifiers[2]].aes_encryption_key,
                ),
            ]
        )
        assert encrypt.call_count == 3
        mock_message_service.send_message.assert_has_calls(
            [
                call(
                    _BitfountMessage(
                        message_type=_BitfountMessageType.TASK_COMPLETE,
                        body=None,  # type: ignore[arg-type] # reason: message body should be empty # noqa: B950
                        recipient=pod_identifiers[0],
                        recipient_mailbox_id=worker_mailboxes[
                            pod_identifiers[0]
                        ].mailbox_id,
                        sender=modeller_name,
                        sender_mailbox_id=modeller_mailbox_id,
                        timestamp=fake_timestamps[0],
                        task_id=task_id,
                    ),
                    already_packed=True,
                ),
                call(
                    _BitfountMessage(
                        message_type=_BitfountMessageType.TASK_COMPLETE,
                        body=None,  # type: ignore[arg-type] # reason: message body should be empty # noqa: B950
                        recipient=pod_identifiers[1],
                        recipient_mailbox_id=worker_mailboxes[
                            pod_identifiers[1]
                        ].mailbox_id,
                        sender=modeller_name,
                        sender_mailbox_id=modeller_mailbox_id,
                        timestamp=fake_timestamps[1],
                        task_id=task_id,
                    ),
                    already_packed=True,
                ),
                call(
                    _BitfountMessage(
                        message_type=_BitfountMessageType.TASK_COMPLETE,
                        body=None,  # type: ignore[arg-type] # reason: message body should be empty # noqa: B950
                        recipient=pod_identifiers[2],
                        recipient_mailbox_id=worker_mailboxes[
                            pod_identifiers[2]
                        ].mailbox_id,
                        sender=modeller_name,
                        sender_mailbox_id=modeller_mailbox_id,
                        timestamp=fake_timestamps[2],
                        task_id=task_id,
                    ),
                    already_packed=True,
                ),
            ]
        )
        assert mock_message_service.send_message.call_count == 3

    async def test_get_oidc_client_ids(
        self,
        mock_aes_decrypter: Mock,
        modeller_mailbox: _ModellerMailbox,
        modeller_mailbox_id: str,
        modeller_name: str,
        opt_task_id: Optional[str],
        patch_poll_for_messages: PollForMessagesPatcher,
        worker_mailbox_details: List[WorkerDetails],
    ) -> None:
        """Test get_oidc_client_ids() extracts needed details."""
        # Create fake task response messages
        oidc_client_ids = [
            _OIDCClientID(f"client_id_{i}").serialize() for i in range(3)
        ]
        oidc_challenges = [
            _BitfountMessage(
                message_type=_BitfountMessageType.OIDC_CHALLENGE,
                body=msgpack.dumps(oidc_client_ids[i]),
                recipient=modeller_name,
                recipient_mailbox_id=modeller_mailbox_id,
                sender=worker_mailbox_details[i].pod_identifier,
                sender_mailbox_id=worker_mailbox_details[i].mailbox_id,
                task_id=opt_task_id,
            )
            for i in range(3)
        ]

        # Assign these to poll_for_messages which is the underlying method we
        # expect to yield messages.
        patch_poll_for_messages(oidc_challenges)

        # Start listening and processing messages
        received_oidc_challenges = await _run_func_and_listen_to_mailbox(
            modeller_mailbox.get_oidc_client_ids(), modeller_mailbox
        )

        # Ensure extracted details match expected
        assert received_oidc_challenges == {
            worker_mailbox_details[i].pod_identifier: _OIDCClientID(
                oidc_client_ids[i]["client_id"]
            )
            for i in range(3)
        }

    async def test_get_oidc_client_ids_throws_error(
        self,
        modeller_mailbox: _ModellerMailbox,
        patch_poll_for_messages: PollForMessagesPatcher,
        rpc_error: RpcError,
    ) -> None:
        """Test Start Task error thrown on error."""
        patch_poll_for_messages([rpc_error])
        with raises(
            BitfountTaskStartError, match="Failed to start task with all pods."
        ):
            await _run_func_and_listen_to_mailbox(
                modeller_mailbox.get_oidc_client_ids(),
                modeller_mailbox,
            )

    async def test_send_oidc_auth_flow_responses(
        self,
        mock_message_service: Mock,
        mocker: MockerFixture,
        modeller_mailbox: _ModellerMailbox,
        modeller_mailbox_id: str,
        modeller_name: str,
        pod_identifiers: List[str],
        task_id: str,
        worker_mailboxes: Dict[str, _WorkerMailboxDetails],
    ) -> None:
        """Test send_oidc_auth_flow_responses() sends expected messages."""
        # Set modeller username
        mock_message_service.username = modeller_name

        # Mock out _send_aes_encrypted_message()
        mock_send_aes = mocker.patch(
            "bitfount.federated.transport.modeller_transport._send_aes_encrypted_message",  # noqa: B950
            autospec=True,
        )

        oidc_details = {
            pod_identifiers[i]: _OIDCAuthFlowResponse(
                auth_code=f"auth_code_{pod_identifiers[i]}",
                code_verifier=f"code_verifier_{pod_identifiers[i]}",
                redirect_uri=f"redirect_uri_{pod_identifiers[i]}",
            )
            for i in range(len(pod_identifiers))
        }

        await modeller_mailbox.send_oidc_auth_flow_responses(oidc_details)

        # Check calls
        mock_send_aes.assert_has_calls(
            [
                call(
                    message={
                        "auth_code": oidc_details[pod_id].auth_code,
                        "code_verifier": oidc_details[pod_id].code_verifier,
                        "redirect_uri": oidc_details[pod_id].redirect_uri,
                    },
                    aes_encryption_key=worker_mailboxes[pod_id].aes_encryption_key,
                    message_service=modeller_mailbox.message_service,
                    message_type=_BitfountMessageType.OIDC_AFC_PKCE_RESPONSE,
                    recipient=worker_mailboxes[pod_id].pod_identifier,
                    recipient_mailbox_id=worker_mailboxes[pod_id].mailbox_id,
                    sender=modeller_name,
                    sender_mailbox_id=modeller_mailbox_id,
                    task_id=task_id,
                )
                for pod_id in pod_identifiers
            ]
        )

    async def test_send_oidc_device_code_responses(
        self,
        mock_message_service: Mock,
        mocker: MockerFixture,
        modeller_mailbox: _ModellerMailbox,
        modeller_mailbox_id: str,
        modeller_name: str,
        pod_identifiers: List[str],
        task_id: str,
        worker_mailboxes: Dict[str, _WorkerMailboxDetails],
    ) -> None:
        """Test send_oidc_auth_flow_responses() sends expected messages."""
        # Set modeller username
        mock_message_service.username = modeller_name

        # Mock out _send_aes_encrypted_message()
        mock_send_aes = mocker.patch(
            "bitfount.federated.transport.modeller_transport._send_aes_encrypted_message",  # noqa: B950
            autospec=True,
        )

        # Create pod device code details for each pod
        now = datetime.now(timezone.utc)
        base_interval = 5
        device_code_details: Dict[str, _PodDeviceCodeDetails] = {
            pod_identifiers[i]: _PodDeviceCodeDetails(
                device_code=f"device_code_{i}",
                expires_at=now + timedelta(hours=i),
                interval=base_interval + i,
            )
            for i in range(len(pod_identifiers))
        }

        await modeller_mailbox.send_oidc_device_code_responses(device_code_details)

        # Check calls
        mock_send_aes.assert_has_calls(
            [
                call(
                    message={
                        "device_code": device_code_details[pod_id].device_code,
                        "expires_at": device_code_details[
                            pod_id
                        ].expires_at.isoformat(),
                        "interval": device_code_details[pod_id].interval,
                    },
                    aes_encryption_key=worker_mailboxes[pod_id].aes_encryption_key,
                    message_service=modeller_mailbox.message_service,
                    message_type=_BitfountMessageType.OIDC_DEVICE_CODE_RESPONSE,
                    recipient=worker_mailboxes[pod_id].pod_identifier,
                    recipient_mailbox_id=worker_mailboxes[pod_id].mailbox_id,
                    sender=modeller_name,
                    sender_mailbox_id=modeller_mailbox_id,
                    task_id=task_id,
                )
                for pod_id in pod_identifiers
            ]
        )

    def test__setup_online_status_handler(
        self, mocker: MockerFixture, modeller_mailbox: _ModellerMailbox
    ) -> None:
        """Test that modeller mailbox registers ONLINE_CHECK message handler."""
        mock_register_handler = mocker.patch.object(
            modeller_mailbox, "register_handler"
        )
        modeller_mailbox._setup_online_status_handler()
        mock_register_handler.assert_called_once()
        assert (
            mock_register_handler.call_args[0][0] == _BitfountMessageType.ONLINE_CHECK
        )

    async def test_modeller_responds_to_online_status_request(
        self,
        caplog: LogCaptureFixture,
        mock_aes_decrypter: Mock,
        mock_message_service: Mock,
        mocker: MockerFixture,
        modeller_mailbox: _ModellerMailbox,
        modeller_mailbox_id: str,
        modeller_name: str,
        online_check_uuid: str,
        opt_task_id: Optional[str],
        patch_poll_for_messages: PollForMessagesPatcher,
        task_id: str,
        worker_mailbox_details: List[WorkerDetails],
    ) -> None:
        """Some pods never send a response to the training request."""
        caplog.set_level("INFO")
        # Pod sends message to modeller checking if they are online
        pod_messages = [
            _BitfountMessage(
                message_type=_BitfountMessageType.ONLINE_CHECK,
                body=online_check_uuid,  # type: ignore[arg-type] # reason: should return unchanged # noqa: B950
                recipient=modeller_name,
                recipient_mailbox_id=modeller_mailbox_id,
                sender=worker_mailbox_details[0].pod_identifier,
                sender_mailbox_id=worker_mailbox_details[0].mailbox_id,
                task_id=opt_task_id,
            ),
        ]

        # For ease of testing, we pretend that the pod has already accepted the task
        modeller_mailbox.accepted_worker_mailboxes = {
            worker_mailbox_details[0].pod_identifier: worker_mailbox_details[0]  # type: ignore[dict-item] # reason: ease of testing # noqa: B950
        }

        mock_send_message = mocker.patch(
            "bitfount.federated.transport.modeller_transport._send_aes_encrypted_message"  # noqa: B950
        )

        # Mock out _BitfountMessage constructor
        mock_bitfount_message_cls = mocker.patch(
            "bitfount.federated.transport.modeller_transport._BitfountMessage"
        )

        # Assign this message to poll_for_messages yield
        patch_poll_for_messages(pod_messages)

        # Start listening and processing messages. We timeout on the responses to
        # simulate non-responding pods. This may need to be increased if the test
        # is failing but should need to be no higher than 5 seconds.
        await _run_func_and_listen_to_mailbox(
            modeller_mailbox.process_task_request_responses(timeout=3), modeller_mailbox
        )

        # Assert that the response was sent to the correct pod and is of the right type
        assert caplog.records[0].levelname == "INFO"
        assert (
            caplog.records[0].message
            == "Informing user1/pod_1 that we are still online."
        )

        # Check correct message sending was used
        mock_send_message.assert_not_called()
        mock_message_service.send_message.assert_called_once_with(
            mock_bitfount_message_cls.return_value, already_packed=True
        )
        mock_bitfount_message_cls.assert_called_once_with(
            body=online_check_uuid,
            message_type=_BitfountMessageType.ONLINE_RESPONSE,
            recipient=worker_mailbox_details[0].pod_identifier,
            recipient_mailbox_id=worker_mailbox_details[0].mailbox_id,
            sender=modeller_mailbox.message_service.username,
            sender_mailbox_id=modeller_mailbox.mailbox_id,
            task_id=task_id,
        )

    async def test_log(
        self, mocker: MockerFixture, modeller_mailbox: _ModellerMailbox
    ) -> None:
        """Test that log message is sent appropriately."""
        mock_log = mocker.patch.object(
            modeller_mailbox, "_send_to_all_pods_aes_encrypt"
        )
        await modeller_mailbox.log({"msg": "message"})
        mock_log.assert_awaited_once_with(
            {"msg": "message"}, _BitfountMessageType.LOG_MESSAGE
        )


@unit_test
class TestModellerTransportFunctions:
    """Tests for functions in modeller_transport module."""

    async def test__parameters_update_handler(
        self,
        mocker: MockerFixture,
        modeller_mailbox: _ModellerMailbox,
        modeller_mailbox_id: str,
        opt_task_id: Optional[str],
        pod_identifiers: List[str],
        worker_mailboxes: Dict[str, _WorkerMailboxDetails],
    ) -> None:
        """Tests that weight updates are populated by handler."""
        # Decrypted & message
        expected_message = {"some parameter": ["some weight list"]}
        decrypt = mocker.patch.object(_BitfountMessage, "decrypt")
        decrypt.return_value = _DecryptedBitfountMessage(
            message_type=_BitfountMessageType.TRAINING_UPDATE,
            body=expected_message,
            recipient=pod_identifiers[0],
            recipient_mailbox_id=worker_mailboxes[pod_identifiers[0]].mailbox_id,
            sender=pod_identifiers[1],
            sender_mailbox_id=modeller_mailbox_id,
            task_id=opt_task_id,
        )

        modeller_mailbox.accepted_worker_mailboxes = worker_mailboxes

        # Message from pod
        expected_encrypted_message = b"encrypted_message"
        message = _BitfountMessage(
            message_type=_BitfountMessageType.TRAINING_UPDATE,
            body=expected_encrypted_message,
            recipient=pod_identifiers[0],
            recipient_mailbox_id=worker_mailboxes[pod_identifiers[0]].mailbox_id,
            sender=pod_identifiers[1],
            sender_mailbox_id=modeller_mailbox_id,
            task_id=opt_task_id,
        )

        weight_updates: Dict[str, _SerializedWeights] = {}
        model_update_handler = _parameter_updates_handler(
            modeller_mailbox, weight_updates
        )
        with _AsyncMultipleResponsesHandler(
            handler=model_update_handler,
            message_types=_BitfountMessageType.TRAINING_UPDATE,
            mailbox=modeller_mailbox,
            responders=modeller_mailbox.accepted_worker_mailboxes.keys(),
        ) as response_handler:
            await response_handler.handler(message)

        assert weight_updates == {pod_identifiers[1]: expected_message}

    async def test__training_metrics_handler(
        self,
        mocker: MockerFixture,
        modeller_mailbox: _ModellerMailbox,
        modeller_mailbox_id: str,
        opt_task_id: Optional[str],
        pod_identifiers: List[str],
        worker_mailboxes: Dict[str, _WorkerMailboxDetails],
    ) -> None:
        """Tests that validation metrics are populated by handler."""
        # Decrypted & message
        expected_message = [{"some parameter": "some metric"}]
        decrypt = mocker.patch.object(_BitfountMessage, "decrypt")
        decrypt.return_value = _DecryptedBitfountMessage(
            message_type=_BitfountMessageType.TRAINING_METRICS,
            body=expected_message,
            recipient=pod_identifiers[0],
            recipient_mailbox_id=worker_mailboxes[pod_identifiers[0]].mailbox_id,
            sender=pod_identifiers[1],
            sender_mailbox_id=modeller_mailbox_id,
            task_id=opt_task_id,
        )

        modeller_mailbox.accepted_worker_mailboxes = worker_mailboxes

        # Message from pod
        expected_encrypted_message = b"encrypted_message"
        message = _BitfountMessage(
            message_type=_BitfountMessageType.TRAINING_METRICS,
            body=expected_encrypted_message,
            recipient=pod_identifiers[0],
            recipient_mailbox_id=worker_mailboxes[pod_identifiers[0]].mailbox_id,
            sender=pod_identifiers[1],
            sender_mailbox_id=modeller_mailbox_id,
            task_id=opt_task_id,
        )

        validation_metrics: List[Mapping[str, str]] = []

        validation_metrics_handler = _training_metrics_handler(
            modeller_mailbox, validation_metrics
        )

        with _AsyncMultipleResponsesHandler(
            handler=validation_metrics_handler,
            message_types=_BitfountMessageType.TRAINING_UPDATE,
            mailbox=modeller_mailbox,
            responders=modeller_mailbox.accepted_worker_mailboxes.keys(),
        ) as response_handler:
            await response_handler.handler(message)
        assert validation_metrics == [expected_message]

    async def test_get_metrics_updates_calls_handler(
        self, mocker: MockerFixture, modeller_mailbox: _ModellerMailbox
    ) -> None:
        """Tests that the appropriate handler is called."""
        mock_handler = Mock()
        mocker.patch(
            "bitfount.federated.transport.modeller_transport._training_metrics_handler",
            mock_handler,
        )

        await _get_training_metrics_from_workers(modeller_mailbox, timeout=1)
        mock_handler.assert_called_once_with(modeller_mailbox, [])

    async def test_get_parameter_updates_calls_handler(
        self, mocker: MockerFixture, modeller_mailbox: _ModellerMailbox
    ) -> None:
        """Tests that the appropriate handler is called."""
        mock_handler = Mock()
        mocker.patch(
            "bitfount.federated.transport.modeller_transport._parameter_updates_handler",  # noqa: B950 ignore line too long
            mock_handler,
        )

        await _get_parameter_updates_from_workers(modeller_mailbox, timeout=1)
        mock_handler.assert_called_once_with(modeller_mailbox, {})

    async def test__public_key_handler(
        self,
        mocker: MockerFixture,
        modeller_mailbox: _ModellerMailbox,
        modeller_mailbox_id: str,
        opt_task_id: Optional[str],
        pod_identifiers: List[str],
        worker_mailboxes: Dict[str, _WorkerMailboxDetails],
    ) -> None:
        """Tests that public key is given by handler."""
        # Decrypted & message
        expected_message = [{"some public_key"}]
        decrypt = mocker.patch.object(_BitfountMessage, "decrypt")
        decrypt.return_value = _DecryptedBitfountMessage(
            message_type=_BitfountMessageType.KEY_EXCHANGE,
            body=expected_message,
            recipient=pod_identifiers[0],
            recipient_mailbox_id=worker_mailboxes[pod_identifiers[0]].mailbox_id,
            sender=pod_identifiers[1],
            sender_mailbox_id=modeller_mailbox_id,
            task_id=opt_task_id,
        )

        modeller_mailbox.accepted_worker_mailboxes = worker_mailboxes

        # Message from pod
        expected_encrypted_message = b"encrypted_message"
        message = _BitfountMessage(
            message_type=_BitfountMessageType.KEY_EXCHANGE,
            body=expected_encrypted_message,
            recipient=pod_identifiers[0],
            recipient_mailbox_id=worker_mailboxes[pod_identifiers[0]].mailbox_id,
            sender=pod_identifiers[1],
            sender_mailbox_id=modeller_mailbox_id,
            task_id=opt_task_id,
        )

        pub_key: List[bytes] = []

        pub_key_handler = _public_key_handler(modeller_mailbox, pub_key)

        with _AsyncMultipleResponsesHandler(
            handler=pub_key_handler,
            message_types=_BitfountMessageType.KEY_EXCHANGE,
            mailbox=modeller_mailbox,
            responders=modeller_mailbox.accepted_worker_mailboxes.keys(),
        ) as response_handler:
            await response_handler.handler(message)
        assert pub_key == [expected_message]

    async def test__get_public_key_calls_handler(
        self, mocker: MockerFixture, modeller_mailbox: _ModellerMailbox
    ) -> None:
        """Tests that the appropriate handler is called."""
        mock_handler = Mock()
        mocker.patch(
            "bitfount.federated.transport.modeller_transport._public_key_handler",  # noqa: B950 ignore line too long
            mock_handler,
        )
        await _get_public_key(modeller_mailbox, timeout=1)
        mock_handler.assert_called_once_with(modeller_mailbox, [])

    async def test__psi_dataset_handler(
        self,
        mocker: MockerFixture,
        modeller_mailbox: _ModellerMailbox,
        modeller_mailbox_id: str,
        opt_task_id: Optional[str],
        pod_identifiers: List[str],
        worker_mailboxes: Dict[str, _WorkerMailboxDetails],
    ) -> None:
        """Tests that psi dataset is given by handler."""
        # Decrypted & message
        expected_message = [{"some data"}]
        decrypt = mocker.patch.object(_BitfountMessage, "decrypt")
        decrypt.return_value = _DecryptedBitfountMessage(
            message_type=_BitfountMessageType.PSI_DATASET,
            body=expected_message,
            recipient=pod_identifiers[0],
            recipient_mailbox_id=worker_mailboxes[pod_identifiers[0]].mailbox_id,
            sender=pod_identifiers[1],
            sender_mailbox_id=modeller_mailbox_id,
            task_id=opt_task_id,
        )

        modeller_mailbox.accepted_worker_mailboxes = worker_mailboxes

        # Message from pod
        expected_encrypted_message = b"encrypted_message"
        message = _BitfountMessage(
            message_type=_BitfountMessageType.PSI_DATASET,
            body=expected_encrypted_message,
            recipient=pod_identifiers[0],
            recipient_mailbox_id=worker_mailboxes[pod_identifiers[0]].mailbox_id,
            sender=pod_identifiers[1],
            sender_mailbox_id=modeller_mailbox_id,
            task_id=opt_task_id,
        )

        psi_dataset: List[Tuple[List[str], List[str]]] = []

        psi_dataset_handler = _psi_dataset_handler(modeller_mailbox, psi_dataset)

        with _AsyncMultipleResponsesHandler(
            handler=psi_dataset_handler,
            message_types=_BitfountMessageType.PSI_DATASET,
            mailbox=modeller_mailbox,
            responders=modeller_mailbox.accepted_worker_mailboxes.keys(),
        ) as response_handler:
            await response_handler.handler(message)
        assert psi_dataset == [expected_message]

    async def test__get_psi_datasets_from_workers_calls_handler(
        self, mocker: MockerFixture, modeller_mailbox: _ModellerMailbox
    ) -> None:
        """Tests that the appropriate handler is called."""
        mock_handler = Mock()
        mocker.patch(
            "bitfount.federated.transport.modeller_transport._psi_dataset_handler",  # noqa: B950 ignore line too long
            mock_handler,
        )
        await _get_psi_datasets_from_workers(modeller_mailbox, timeout=1)
        mock_handler.assert_called_once_with(modeller_mailbox, [])



"""Fixtures for the web_utils module."""
import asyncio
import time
from typing import Optional
from unittest.mock import AsyncMock, Mock

from pytest import fixture
from pytest_mock import MockerFixture

from tests.utils import PytestRequest
from tests.utils.helper import end_to_end_mocks_test, integration_test, unit_test


@fixture(autouse=True)
def remove_web_retry_backoff_sleep(
    mocker: MockerFixture, request: PytestRequest
) -> Optional[Mock]:
    """Removes backoff sleep for the auto_retry functionality.

    As we don't want to slow down tests unnecessarily, this fixture mocks out
    the sleep that would normally occur in the automatic retry of web interactions.
    """
    # Only mock it out for tests that don't actually perform web interactions
    # i.e. not e2e or tutorial tests
    if any(
        mark.name in request.keywords
        for mark in (unit_test, integration_test, end_to_end_mocks_test)
    ):
        # Need to patch entire imported module but only want to mock out `.sleep`
        # calls so set that explicitly to a Mock that isn't wrapping
        mock_time: Mock = mocker.patch("bitfount.web_utils.time", wraps=time)
        mock_time.sleep = Mock()
        return mock_time.sleep
    else:
        return None


@fixture(autouse=True)
def remove_async_web_retry_backoff_sleep(
    mocker: MockerFixture, request: PytestRequest
) -> Optional[AsyncMock]:
    """Removes backoff sleep for the async auto_retry functionality.

    As we don't want to slow down tests unnecessarily, this fixture mocks out
    the sleep that would normally occur in the automatic retry of web interactions.
    """
    # Only mock it out for tests that don't actually perform web interactions
    # i.e. not e2e or tutorial tests
    if any(
        mark.name in request.keywords
        for mark in (unit_test, integration_test, end_to_end_mocks_test)
    ):
        # Need to patch entire imported module but only want to mock out `.sleep`
        # calls so set that explicitly to a Mock that isn't wrapping
        mock_async: Mock = mocker.patch("bitfount.web_utils.asyncio", wraps=asyncio)
        mock_async.sleep = AsyncMock()
        return mock_async.sleep
    else:
        return None



"""Provides methods for interacting with the Bitfount Web frontend via Selenium."""
from contextlib import contextmanager
import logging
from pathlib import Path
import threading
import time
from typing import Generator

import chromedriver_autoinstaller
from requests import HTTPError
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from bitfount.config import _STAGING_ENVIRONMENT, _get_environment
from bitfount.hub.api import (
    _STAGING_AM_URL,
    PRODUCTION_AM_URL,
    BitfountAM,
    BitfountSession,
)
from bitfount.hub.authentication_flow import (
    _HUB_API_IDENTIFIER,
    _SCOPES,
    _STAGING_AUTH_DOMAIN,
    _STAGING_CLIENT_ID,
    _get_auth_environment,
)

SCREENSHOT_DIRECTORY: Path = Path("selenium-screenshots")
IMPLICIT_WAIT_TIME = 8  # seconds; high value due to slow GitHub runner

logger = logging.getLogger(__name__)


@contextmanager
def webdriver_factory(
    wait_time: int = IMPLICIT_WAIT_TIME,
) -> Generator[WebDriver, None, None]:
    """Create a Selenium webdriver instance."""
    chromedriver_autoinstaller.install()
    chrome_options = webdriver.ChromeOptions()
    chrome_options.add_argument("--headless")
    chrome_options.add_argument("--window-size=1920,1080")
    with webdriver.Chrome(options=chrome_options) as driver:
        driver.implicitly_wait(wait_time)
        yield driver


def save_screenshot(file_name: str, driver: WebDriver) -> None:
    """Saves a screenshot of the Selenium driver view."""
    SCREENSHOT_DIRECTORY.mkdir(parents=True, exist_ok=True)
    file_path = SCREENSHOT_DIRECTORY / f"{file_name}_{str(int(time.time()))}.png"
    file_path = file_path.resolve()
    driver.get_screenshot_as_file(str(file_path))
    logger.error(f"Screenshot saved to {file_path}")


def _oauth_sign_in(username: str, password: str, driver: WebDriver) -> None:
    """Sign in to the oauth form using the supplied username and password."""
    logger.info(f"OAuth sign in for {username} at {driver.current_url}")
    try:
        # Supply login details to Auth0 login panel
        driver.find_element_by_css_selector("input[name='username']").send_keys(
            username
        )
        driver.find_element_by_css_selector("input[name='password']").send_keys(
            password
        )
        driver.find_element_by_css_selector("button[name='action']").click()
    except Exception as e:
        logger.error("Exception encountered whilst signing in.")
        save_screenshot("perform_login", driver)
        raise e


class WebdriverBitfountSession(BitfountSession):
    """A BitfountSession implementation using Selenium Webdriver.

    An implementation of BitfountSession which uses a Selenium Webdriver to approve
    the session access.
    """

    def __init__(
        self,
        password: str,
        # Below are the inherited args
        username: str,
        auth_domain: str = _STAGING_AUTH_DOMAIN,
        client_id: str = _STAGING_CLIENT_ID,
        scopes: str = _SCOPES,
        audience: str = _HUB_API_IDENTIFIER,
    ):
        logger.info(f"Webdriver is using {auth_domain} for user {username}")
        super().__init__(
            username=username,
            auth_domain=auth_domain,
            client_id=client_id,
            scopes=scopes,
            audience=audience,
        )
        self.password = password

    def _do_verification(self, user_code: str, verification_uri: str) -> None:
        """Opens web browser for verification."""
        with webdriver_factory() as driver:
            try:
                # Load Auth0 location
                driver.get(verification_uri)

                # Extract code
                secure_code_field = driver.find_element_by_css_selector(
                    "input[aria-label='Secure code']"
                )
                secure_code = secure_code_field.get_attribute("value")
                if secure_code != user_code:
                    raise ValueError(
                        f"Secure codes differ: {user_code} != {secure_code}"
                    )

                # If code ok, click button
                driver.find_element_by_css_selector("button[value='confirm']").click()
                # Perform login
                _oauth_sign_in(self._username, self.password, driver)
                # Wait for page to load
                driver.find_element_by_xpath(
                    "//p[text()='Your device is now connected.']"
                )
            except Exception as e:
                logger.error(
                    f"Exception encountered whilst attempting to perform oauth "
                    f"verification at {verification_uri}."
                )
                save_screenshot("do_verification", driver)
                raise e


class ExtendedBitfountAM(BitfountAM):
    """Extends BitfountAM with methods relevant to testing."""

    def __init__(self, session: WebdriverBitfountSession, access_manager_url: str):
        logger.info(f"Using {access_manager_url} for access manager")
        super().__init__(session, access_manager_url)

    def grant_proactive_access(
        self, pod_id: str, user_to_grant: str, role: str
    ) -> None:
        """Sets a pod to grant proactive access to the username specified."""
        response = self.session.post(
            f"{self.access_manager_url}/api/casbin",
            timeout=10,
            json={
                "podIdentifier": pod_id,
                "grantee": user_to_grant,
                "role": role,
            },
        )

        if response.status_code not in (200, 201):
            raise HTTPError(
                f"Unexpected response ({response.status_code}): {response.text}"
            )


def get_bitfount_session(
    username: str,
    password: str,
    token_dir: Path,
) -> WebdriverBitfountSession:
    """Creates and returns a WebdriverBitfountSession."""
    bf_env = _get_auth_environment()
    session = WebdriverBitfountSession(
        username=username,
        password=password,
        auth_domain=bf_env.auth_domain,
        client_id=bf_env.client_id,
    )
    session.user_storage_path = token_dir / username
    session.token_file = session.user_storage_path / ".token"
    return session


def oidc_flow(
    url: str,
    username: str,
    password: str,
) -> None:
    """Opens provided oidc url and logs in before closing browser.

    This is run in a separate thread so that the Modeller can respond to the challenges
    from the pods. Otherwise this ends up blocking the Modeller.

    Args:
        url (str): the oidc authentication url to open. This is provided at run-time.
        username (str): the username of the user to log in as
        password (str): the password of the user
    """

    def execute_oidc_flow() -> None:
        with webdriver_factory() as driver:
            try:
                logger.warning("opening url in browser")
                driver.get(url)

                # Wait for a maximum of 30 seconds for the OIDC Confirmation to appear
                # The webdriver won't wait for the full 30 seconds if the element is
                # located before this time.
                WebDriverWait(driver, 30).until(
                    EC.visibility_of_element_located(
                        (By.XPATH, '//button[@value="confirm"]')
                    )
                )
                driver.find_element_by_css_selector("button[value='confirm']").click()

                _oauth_sign_in(username, password, driver)

                # Wait for a maximum of 30 seconds for the device to connect
                WebDriverWait(driver, 30).until(
                    EC.visibility_of_element_located(
                        (By.XPATH, "//p[text()='Your device is now connected.']")
                    )
                )

            except Exception as e:
                logger.error(
                    f"Exception encountered whilst attempting to perform oidc auth "
                    f"verification at {url}."
                )
                save_screenshot("do_oidc_verification", driver)
                raise e

    oidc_thread = threading.Thread(target=execute_oidc_flow, name="oidc")
    oidc_thread.start()


def grant_proactive_access(
    modeller_username: str,
    pod_id: str,
    role: str,
    pod_session: WebdriverBitfountSession,
) -> None:
    """Grants proactive access to a pod for a given modeller."""
    bf_env = _get_environment()
    if bf_env is _STAGING_ENVIRONMENT:
        am_url = _STAGING_AM_URL
    else:
        am_url = PRODUCTION_AM_URL

    am = ExtendedBitfountAM(pod_session, am_url)
    am.grant_proactive_access(
        pod_id=pod_id,
        user_to_grant=modeller_username,
        role=role,
    )
